"""
==================================
Set Axis Range When Plotting a Map
==================================

In this example we are going to look at how to set the axes
range using Matplotlib's ``set_xlim`` and ``set_ylim`` when plotting a
Map with WCSAxes.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.data.sample
import sunpy.map

###############################################################################
# Lets start by creating a Map from the sample data.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)

###############################################################################
# Now lets say for example we are only interested in plotting a certain region
# of this Map. One way this could be done is to create a submap over the region
# of interest and then plotting that. Another useful way is to set the axes
# range over which to plot using Matplotlib's
# `~matplotlib.axes.Axes.set_xlim` and `~matplotlib.axes.Axes.set_ylim` functionality.
# The axes that Matplotlib uses is in pixel coordinates (e.g. of image data array)
# rather than world coordinates (e.g. in arcsecs) so we need to define our limits that
# are passed to `~matplotlib.axes.Axes.set_xlim`, `~matplotlib.axes.Axes.set_ylim` to pixel coordinates.
# We can define our limits we want to use in world coordinates and then work out what pixel
# coordinates these correspond to.
# Lets choose x-limits and y-limits in arcsecs that we are interested in.

xlims_world = [500, 1100]*u.arcsec
ylims_world = [-800, 0]*u.arcsec

###############################################################################
# We can then covert these into a SkyCoord which can be passed to :func:`~sunpy.map.GenericMap.world_to_pixel` to
# determine which pixel coordinates these represent on the Map.

world_coords = SkyCoord(Tx=xlims_world, Ty=ylims_world, frame=aia_map.coordinate_frame)
pixel_coords = aia_map.world_to_pixel(world_coords)

# we can then pull out the x and y values of these limits.
xlims_pixel = pixel_coords.x.value
ylims_pixel = pixel_coords.y.value

###############################################################################
# We can now plot this Map and then use the x_lims_pixel and y_lims_pixel to set
# the range of the axes for which to plot.

fig = plt.figure()
ax = fig.add_subplot(projection=aia_map)
aia_map.plot(axes=ax, clip_interval=(1, 99.9)*u.percent)
ax.set_xlim(xlims_pixel)
ax.set_ylim(ylims_pixel)

plt.show()



import sunpy.timeseries
from sunpy.data.test import get_test_filepath

fermi_gbm_filepath = get_test_filepath('gbm.fits')


def test_implicit_fermi_gbm():
    # Test a GBMSummary TimeSeries
    ts_gbm = sunpy.timeseries.TimeSeries(fermi_gbm_filepath)
    assert isinstance(ts_gbm, sunpy.timeseries.sources.fermi_gbm.GBMSummaryTimeSeries)


def test_fermi_gbm():
    # Test a GBMSummary TimeSeries
    ts_gbm = sunpy.timeseries.TimeSeries(fermi_gbm_filepath, source='GBMSummary')
    assert isinstance(ts_gbm, sunpy.timeseries.sources.fermi_gbm.GBMSummaryTimeSeries)



from sunpy.net.dataretriever import GenericClient

__all__ = ['GONGClient']


class GONGClient(GenericClient):
    """
    Provides access to the Magnetogram products of NSO-GONG synoptic Maps.

    Searches data hosted by the `National Solar Observatory <gong2.nso.edu/oQR/zqs/>`__

    Examples
    --------
    >>> from sunpy.net import Fido, attrs as a
    >>> results = Fido.search(a.Time("2019/12/31 22:00", "2020/1/1 02:00"),
    ...                       a.Instrument('GONG'))  #doctest: +REMOTE_DATA
    >>> results  #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    5 Results from the GONGClient:
    Source: https://gong2.nso.edu/oQR/zqs
    <BLANKLINE>
           Start Time               End Time        ... Provider ExtentType
    ----------------------- ----------------------- ... -------- ----------
    2019-12-31 22:14:00.000 2019-12-31 22:14:59.999 ...      NSO   SYNOPTIC
    2019-12-31 23:04:00.000 2019-12-31 23:04:59.999 ...      NSO   SYNOPTIC
    2019-12-31 23:54:00.000 2019-12-31 23:54:59.999 ...      NSO   SYNOPTIC
    2020-01-01 00:14:00.000 2020-01-01 00:14:59.999 ...      NSO   SYNOPTIC
    2020-01-01 01:14:00.000 2020-01-01 01:14:59.999 ...      NSO   SYNOPTIC
    <BLANKLINE>
    <BLANKLINE>
    """
    baseurl = (r'https://gong2.nso.edu/oQR/zqs/%Y%m/mrzqs%y%m%d/mrzqs%y%m%dt%H%Mc'
               r'(\d){4}_(\d){3}\.fits.gz')
    pattern = '{}/zqs/{year:4d}{month:2d}/mrzqs{:4d}{day:2d}/mrzqs{:6d}t{hour:2d}{minute:2d}c{}'

    @property
    def info_url(self):
        return 'https://gong2.nso.edu/oQR/zqs'

    @classmethod
    def register_values(cls):
        from sunpy.net import attrs
        adict = {attrs.Instrument: [("GONG", "Global Oscillation Network Group.")],
                 attrs.Physobs: [("LOS_MAGNETIC_FIELD", "Line of sight magnetic field")],
                 attrs.Provider: [('NSO', 'National Solar Observatory.')],
                 attrs.ExtentType: [("SYNOPTIC", "Coverage of a complete solar rotation synthesized over time")]}
        return adict



import warnings

import pytest

from sunpy.util.decorators import deprecated, get_removal_version
from sunpy.util.exceptions import SunpyDeprecationWarning, SunpyPendingDeprecationWarning


def test_removal_version_since_lts():
    major, minor = get_removal_version('2.0')
    assert major == 2
    assert minor == 1


def test_removal_version_not_since_lts():
    major, minor = get_removal_version('2.1')
    assert major == 3
    assert minor == 1


@pytest.mark.parametrize(
    'since,pending,warning,message,warning_message',
    [
        ('2.0', False, SunpyDeprecationWarning, '',
         'The foo function is deprecated and may be removed in version 2.1.'),
        ('2.1', False, SunpyDeprecationWarning, '',
         'The foo function is deprecated and may be removed in version 3.1.'),
        ('2.0', True, SunpyPendingDeprecationWarning, '',
         'The foo function will be deprecated in version 2.0.'),
        ('2.0', False, SunpyDeprecationWarning,
         'Custom deprecation message', 'Custom deprecation message'),
    ]
)
def test_deprecated_warning_message(since, pending, warning, message, warning_message):
    @deprecated(since, pending=pending, message=message)
    def foo():
        pass
    with pytest.warns(warning, match=warning_message):
        warnings.simplefilter('always')
        foo()



from textwrap import dedent

import numpy as np
import pytest

import astropy.units as u
from astropy.io import fits

from sunpy.map import Map
from sunpy.map.sources.hinode import SOTMap
from sunpy.util.exceptions import SunpyMetadataWarning


@pytest.fixture
def sot():
    raw_header = dedent("""\
        SIMPLE  =                    T / created Wed Oct 21 17:17:46 2015
        BITPIX  =                   16
        NAXIS   =                    2
        NAXIS1  =                 2048
        NAXIS2  =                 1024
        EXTEND  =                    T
        DATE    = '2015-10-21T17:17:46.000'        /creation date
        DATE_RF0= '2015-10-21T17:17:46.000'        /creation date
        TELESCOP= 'HINODE'
        MDP_CLK =            361702812
        FILEORIG= '2015_1021_171239.sci'
        MDPCTREF=            361701639
        CTREF   =           1801875595
        CTRATE  =              584.000
        TIMEERR =                    0
        EXP0    =             0.123088
        OBT_TIME=            361701866
        OBT_END =            361701929
        DATE_OBS= '2015-10-13T23:13:44.601'
        TIME-OBS= '23:13:44.601'
        CTIME   = 'Tue Oct 13 23:13:44 2015'
        DATE_END= '2015-10-13T23:13:44.724'
        TIMESPAN=             0.123088
        TIMESYS = 'UTC'
        INSTRUME= 'SOT/WB'
        ORIGIN  = 'JAXA/ISAS, SIRIUS'
        DATA_LEV=                    0
        ORIG_RF0= 'JAXA/ISAS, SIRIUS'
        VER_RF0 = '1.66'
        PROG_VER=                  609
        SEQN_VER=                  919
        PARM_VER=                  628
        PROG_NO =                   17
        SUBR_NO =                    1
        SEQN_NO =                   27
        MAIN_CNT=                    1
        MAIN_RPT=                    1
        MAIN_POS=                    1
        SUBR_CNT=                    1
        SUBR_RPT=                    2
        SUBR_POS=                    1
        SEQN_CNT=                    1
        SEQN_RPT=                    1
        SEQN_POS=                    5
        OBSTITLE= ' '
        TARGET  = ' '
        SCI_OBJ = ' '
        SCI_OBS = 'TBD'
        OBS_DEC = ' '
        JOIN_SB = ' '
        OBS_NUM =                    0
        JOP_ID  =                    0
        NOAA_NUM=                    0
        OBSERVER= '  '
        PLANNER = '  '
        TOHBANS = '  '
        DATATYPE= 'SCI'
        FLFLG   = 'NON'
        OBS_MODE= 'QT'
        SAA     = 'OUT'
        HLZ     = 'OUT'
        OBS_ID  =                    1
        GEN_ID  =                    1
        FRM_ID  =                    2
        WAVEID  =                    2
        OBS_TYPE= 'FG (simple)'
        MACROID =                13348
        XSCALE  =             0.108960
        YSCALE  =             0.108960
        FGXOFF  =                    0
        FGYOFF  =                    0
        FGCCDIX0=                    0
        FGCCDIX1=                 4095
        FGCCDIY0=                    0
        FGCCDIY1=                 2047
        CRPIX1  =              1024.50
        CRPIX2  =              512.500
        SC_ATTX =             -15.8358
        SC_ATTY =              19.2347
        CRVAL1  =             -15.8358
        CRVAL2  =              19.2347
        CDELT1  =             0.108960
        CDELT2  =             0.108960
        CUNIT1  = 'arcsec'
        CUNIT2  = 'arcsec'
        CTYPE1  = 'Solar-X'
        CTYPE2  = 'Solar-Y'
        SAT_ROT =              0.00000
        INST_ROT=             0.412000
        CROTA1  =             0.412000
        CROTA2  =             0.412000
        XCEN    =             -15.8358
        YCEN    =              19.2347
        FOVX    =              223.150
        FOVY    =              111.575
        TR_MODE = 'FIX'
        FGBINX  =                    1
        FGBINY  =                    1
        EXPTIME =             0.122880
        WAVE    = 'Ca II H line'
        DARKFLAG=                    0
        BITCOMP1=                    6
        IMGCOMP1=                    7
        QTABLE1 =                    2
        BITCOMP2=                    1
        IMGCOMP2=                    7
        QTABLE2 =                    1
        PCK_SN0 =             42477799
        PCK_SN1 =             42477831
        NUM_PCKS=                   33
        FGMODE  = 'shuttered'
        FGNINT  =                    1
        ROILOOP =                    0
        NROILOOP=                    0
        CTSERVO =                    1
        CTMESTAT=                36864
        CTMEX   =                 -801
        CTMEY   =                  329
        CTMODE  =                   35
        T_SPCCD =             -43.4724
        T_FGCCD =             -32.2096
        T_CTCCD =              26.6964
        T_SPCEB =             -6.63403
        T_FGCEB =              1.65425
        T_CTCEB =              2.92207
        MASK    =                   22
        WBFW    =                  118
        WEDGE   =                   21
        NBFW    =                  118
        TF1     =                  132
        TF2     =                  133
        TF3     =                  137
        TF4     =                  129
        TF5     =                  106
        TF6     =                   54
        TF7     =                   17
        TF8     =                  154
        SLITENC =                 1991
        FOCUS   =                 2035
        WBEXP   =                   50
        NBEXP   =                  199
        WAVEOFF =                    0
        ROISTART=                    0
        ROISTOP =                 1025
        DOPVUSED=                 -110
        CAMGAIN =                    2
        CAMDACA =                    8
        CAMDACB =                    8
        CAMPSUM =                    2
        CAMSSUM =                    2
        CAMAMP  =                    0
        CAMSCLK =                    0
        PMUDELAY=                  128
        BITCVER1=                45094
        ACHFVER1=                40961
        DCHFVER1=                53249
        QTABVER1=                57365
        BYTECNTI=               401402
        PIXCNTI =              2097152
        BITSPPI =              1.53123
        BYTECNTQ=                    0
        PIXCNTQ =                    0
        BITSPPQ =              0.00000
        PERCENTD=              100.000
        """)
    header = fits.Header.fromstring(raw_header, sep='\n')
    data = np.random.rand(header['naxis1'], header['naxis2'])
    return Map(data, header)


# SOT Tests
def test_fitstoSOT(sot):
    """Tests the creation of SOTMap using FITS."""
    assert isinstance(sot, SOTMap)


def test_is_datasource_for(sot):
    """Test the is_datasource_for method of SOTMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert sot.is_datasource_for(sot.data, sot.meta)


def test_observatory(sot):
    """Tests the observatory property of the SOTMap object."""
    assert sot.observatory == "Hinode"


def test_measurement(sot):
    """Tests the measurement property of the SOTMap object."""
    assert sot.measurement is None


def test_instruments(sot):
    """Tests the Instruments object of SOTMap."""
    assert (sot.Instruments == ['SOT/WB',
                                'SOT/NB', 'SOT/SP', 'SOT/CT'])


def test_waves(sot):
    """Tests the Waves object of SOTMap."""
    assert (sot.Waves == ['6302A', 'BFI no move',
                          'CN bandhead 3883', 'Ca II H line',
                          'G band 4305', 'NFI no move', 'TF Fe I 6302',
                          'TF Mg I 5172', 'TF Na I 5896',
                          'blue cont 4504', 'green cont 5550',
                          'red cont 6684'])


def test_obstype(sot):
    """Tests the Observation_Type object of SOTMap."""
    assert (sot.Observation_Type == ['FG (simple)',
                                     'FG focus scan', 'FG shuttered I and V',
                                     'FG shutterless I and V', 'FG shutterless I and V with 0.2s intervals',
                                     'FG shutterless Stokes', 'SP IQUV 4D array'])


def test_wcs(sot):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    with pytest.warns(SunpyMetadataWarning, match='assuming Earth-based observer'):
        sot.pixel_to_world(0*u.pix, 0*u.pix)



# Author: Simon Liedtke <liedtke.simon@googlemail.com>
#
# This module was developed with funding provided by
# the Google Summer of Code (2013).

import fnmatch

import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from sunpy.database.commands import (
    AddEntry,
    AddTag,
    CommandManager,
    CompositeOperation,
    EditEntry,
    EmptyCommandStackError,
    NonRemovableTagError,
    NoSuchEntryError,
    RemoveEntry,
    RemoveTag,
)
from sunpy.database.tables import DatabaseEntry, Tag


@pytest.fixture
def session():
    # always create an in-memory database with its own new table in each test
    engine = create_engine('sqlite:///:memory:')
    Session = sessionmaker()
    DatabaseEntry.metadata.create_all(bind=engine)
    return Session(bind=engine)


@pytest.fixture
def command_manager():
    return CommandManager()


def test_add_entry_repr(session):
    entry = DatabaseEntry(id=5)
    repr_result = repr(AddEntry(session, entry))
    expected_repr_result = (
        '<AddEntry('
        'session <sqlalchemy.orm.session.Session object at *>, '
        'entry id 5)>'.format(id(session)))
    assert fnmatch.fnmatch(repr_result, expected_repr_result)


def test_add_entry(session):
    assert not session.new
    entry = DatabaseEntry()
    AddEntry(session, entry)()
    assert len(session.new) == 1
    assert entry.id is None
    session.commit()
    assert not session.new
    assert entry.id == 1


def test_add_entry_undo(session):
    entry = DatabaseEntry()
    cmd = AddEntry(session, entry)
    cmd()
    assert session.query(DatabaseEntry).count() == 1
    assert entry.id == 1
    cmd.undo()
    assert entry in session.deleted
    assert session.query(DatabaseEntry).count() == 0


def test_add_removed_entry(session):
    entry = DatabaseEntry()
    AddEntry(session, entry)()
    session.commit()
    RemoveEntry(session, entry)()
    session.commit()
    AddEntry(session, entry)()
    session.commit()
    assert not session.new
    assert entry.id == 1


def test_add_entry_undo_precommit(session):
    entry = DatabaseEntry()
    cmd = AddEntry(session, entry)
    cmd()
    cmd.undo()
    session.commit()
    assert session.query(DatabaseEntry).count() == 0


def test_edit_entry_repr():
    entry = DatabaseEntry(id=7)
    expected_repr_result = "<EditEntry(kwargs {'foo': 'bar'}, entry id 7)>"
    assert fnmatch.fnmatch(repr(EditEntry(entry, foo='bar')), expected_repr_result)


def test_edit_entry_invalid(session):
    with pytest.raises(ValueError):
        EditEntry(DatabaseEntry())


def test_edit_entry(session):
    entry = DatabaseEntry()
    session.add(entry)
    session.commit()
    assert entry.id == 1
    EditEntry(entry, id=42)()
    assert entry.id == 42


def test_edit_entry_undo(session):
    entry = DatabaseEntry()
    session.add(entry)
    session.commit()
    cmd = EditEntry(entry, id=42)
    cmd()
    session.commit()
    assert entry.id == 42
    cmd.undo()
    session.commit()
    assert entry.id == 1


def test_remove_entry_repr(session):
    entry = DatabaseEntry(id=3)
    expected_repr_result = (
        '<RemoveEntry('
        'session <sqlalchemy.orm.session.Session object at *>, '
        'entry DatabaseEntry(id=3))>'.format(id(session)))
    assert fnmatch.fnmatch(repr(RemoveEntry(session, entry)), expected_repr_result)


def test_remove_existing_entry(session):
    entry = DatabaseEntry()
    session.add(entry)
    assert session.query(DatabaseEntry).count() == 1
    assert entry.id == 1
    RemoveEntry(session, entry)()
    assert entry in session.deleted
    assert session.query(DatabaseEntry).count() == 0


def test_remove_nonexisting_entry(session):
    with pytest.raises(NoSuchEntryError):
        RemoveEntry(session, DatabaseEntry())()


def test_remove_entry_undo(session):
    entry = DatabaseEntry()
    session.add(entry)
    cmd = RemoveEntry(session, entry)
    session.commit()
    cmd()
    assert session.query(DatabaseEntry).count() == 0
    cmd.undo()
    assert session.query(DatabaseEntry).count() == 1


def test_add_tag_repr(session):
    entry = DatabaseEntry(id=12)
    tag = Tag('spam')
    expected_repr_result = (
        "<AddTag("
        "tag 'spam', "
        "session <sqlalchemy.orm.session.Session object at *>, "
        "entry id 12)>".format(id(session)))
    assert fnmatch.fnmatch(repr(AddTag(session, entry, tag)), expected_repr_result)


def test_add_tag(session):
    tag = Tag('tag')
    entry = DatabaseEntry()
    assert entry.tags == []
    cmd = AddTag(session, entry, tag)
    cmd()
    assert tag in entry.tags


def test_add_removed_tag(session):
    entry = DatabaseEntry()
    tag = Tag('tag')
    entry.tags.append(tag)
    session.add(tag)
    session.commit()
    session.delete(tag)
    AddTag(session, entry, tag)()
    assert tag in entry.tags


def test_add_tag_undo_unsaved_entry(session):
    tag = Tag('tag')
    entry = DatabaseEntry()
    cmd = AddTag(session, entry, tag)
    cmd()
    cmd.undo()
    assert entry.tags == []
    cmd()
    assert tag in entry.tags


def test_remove_tag_repr(session):
    entry = DatabaseEntry(id=8)
    tag = Tag('foo')
    expected_repr_result = (
        "<RemoveTag("
        "tag 'foo', "
        "session <sqlalchemy.orm.session.Session object at *>, "
        "entry id 8)>".format(id(session)))
    assert fnmatch.fnmatch(repr(RemoveTag(session, entry, tag)), expected_repr_result)


def test_remove_nonexisting_tag(session):
    cmd = RemoveTag(session, DatabaseEntry(), Tag('tag'))
    with pytest.raises(NonRemovableTagError):
        cmd()


def test_remove_tag_undo(session):
    tag = Tag('tag')
    entry = DatabaseEntry()
    entry.tags.append(tag)
    session.add(entry)
    session.commit()
    assert tag in entry.tags
    cmd = RemoveTag(session, entry, tag)
    cmd()
    assert tag not in entry.tags
    cmd.undo()
    assert tag in entry.tags


def test_cmd_manager_pop_undo_cmd(session, command_manager):
    cmd = AddEntry(session, DatabaseEntry())
    command_manager.do(cmd)
    popped_cmd = command_manager.pop_undo_command()
    assert popped_cmd == cmd


def test_cmd_manager_pop_undo_cmd_empty_stack(command_manager):
    with pytest.raises(EmptyCommandStackError):
        command_manager.pop_undo_command()


def test_cmd_manager_pop_redo_cmd(command_manager):
    with pytest.raises(EmptyCommandStackError):
        command_manager.pop_redo_command()


def test_cmd_manager_pop_redo_cmd_empty_stack(session, command_manager):
    cmd = AddEntry(session, DatabaseEntry())
    command_manager.do(cmd)
    command_manager.undo()
    popped_cmd = command_manager.pop_redo_command()
    assert popped_cmd == cmd


def test_cmd_manager_redo_stack_empty_after_call(session, command_manager):
    command_manager.do(AddEntry(session, DatabaseEntry()))
    command_manager.do(AddEntry(session, DatabaseEntry()))
    assert len(command_manager.undo_commands) == 2
    session.commit()
    command_manager.undo(2)
    assert not command_manager.undo_commands
    assert len(command_manager.redo_commands) == 2
    command_manager.do(AddEntry(session, DatabaseEntry()))
    assert not command_manager.redo_commands


def test_cmd_manager_redo(session, command_manager):
    assert command_manager.undo_commands == []
    assert command_manager.redo_commands == []
    command_manager.do(AddEntry(session, DatabaseEntry()))
    command_manager.do(AddEntry(session, DatabaseEntry()))
    assert len(command_manager.undo_commands) == 2
    assert command_manager.redo_commands == []
    session.commit()
    command_manager.undo(2)
    assert command_manager.undo_commands == []
    assert len(command_manager.redo_commands) == 2
    command_manager.redo(2)
    assert len(command_manager.undo_commands) == 2
    assert command_manager.redo_commands == []


def test_undo_redo_multiple_cmds_at_once(session, command_manager):
    assert command_manager.undo_commands == []
    command_manager.do(CompositeOperation([
        AddEntry(session, DatabaseEntry()),
        AddEntry(session, DatabaseEntry()),
        AddEntry(session, DatabaseEntry())]))
    assert len(command_manager.undo_commands) == 1
    assert session.query(DatabaseEntry).count() == 3
    command_manager.undo()
    assert command_manager.undo_commands == []
    assert session.query(DatabaseEntry).count() == 0
    command_manager.redo()
    assert command_manager.redo_commands == []
    assert session.query(DatabaseEntry).count() == 3



"""Yohkoh SXT Map subclass definitions"""

__author__ = "Jack Ireland"
__email__ = "jack.ireland@nasa.gov"

import numpy as np

from astropy.visualization import PowerStretch
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy.map import GenericMap
from sunpy.map.sources.source_type import source_stretch
from sunpy.sun import constants

__all__ = ['SXTMap']


class SXTMap(GenericMap):
    """Yohkoh SXT Image Map

    The Yohkoh Soft X-ray Telescope (SXT) the full solar disk
    (42 x 42 arcminutes)in the 0.25 - 4.0 keV range.
    It consists of a glancing incidence mirror and a CCD sensor and
    used thin metallic filters to acquire images in restricted
    portions of its energy range. SXT could resolve features down to 2.5
    arcseconds. Information about the temperature and density of the plasma
    emitting the observed x-rays was obtained by comparing images acquired with
    the different filters. Images could be obtained every 2 to 8 seconds.
    Smaller images with a single filter could be obtained as frequently as
    once every 0.5 seconds.

    Yohkoh was launched on 30 August 1991 and ceased operations on
    14 December 2001.

    References
    ----------
    * `Yohkoh Mission Page <http://solar.physics.montana.edu/sxt/>`_
    * `Fits header reference <http://proba2.oma.be/data/SWAP/level0>`_
    * `Yohkoh Analysis Guide <http://ylstone.physics.montana.edu/ylegacy/yag.html>`_
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self.plot_settings['cmap'] = 'yohkohsxt' + self.measurement[0:2].lower()
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, PowerStretch(0.5)), clip=False)

    @property
    def observatory(self):
        return "Yohkoh"

    @property
    def detector(self):
        return "SXT"

    @property
    def dsun(self):
        """
        For Yohkoh Maps, DSUN_OBS is not always defined. In this case the
        SOLAR_R keyword is used to calculate dsun.
        """

        # 2012/12/19 - the SXT headers do not have a value of the distance from
        # the spacecraft to the center of the Sun.  The FITS keyword 'DSUN_OBS'
        # appears to refer to the observed diameter of the Sun.  Until such
        # time as that is calculated and properly included in the file, we will
        # use simple trigonometry to calculate the distance of the center of
        # the Sun from the spacecraft.  Note that the small angle approximation
        # is used, and the solar radius stored in SXT FITS files is in arcseconds.

        if 'solar_r' in self.meta:
            dsun = constants.radius / (np.deg2rad(self.meta['solar_r'] / 3600.0))
        else:
            dsun = constants.au
        return self.meta.get('dsun_obs', dsun)

    @property
    def measurement(self):
        s = self.meta.get('wavelnth', '')
        if s == 'Al.1':
            s = 'Al01'
        elif s.lower() == 'open':
            s = 'white-light'
        return s

    @property
    def wavelength(self):
        """
        Returns `None`, as SXT is a broadband imager.
        """

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an SXT image"""
        return header.get('instrume') == 'SXT'



from pathlib import Path
from collections import OrderedDict

import numpy as np

import sunpy
from sunpy import config
from sunpy.net import attrs as a
from sunpy.net.attr import SimpleAttr
from sunpy.net.base_client import BaseClient, QueryResponseRow, QueryResponseTable
from sunpy.net.scraper import Scraper, get_timerange_from_exdict
from sunpy.time import TimeRange
from sunpy.util.parfive_helpers import Downloader

TIME_FORMAT = config.get("general", "time_format")

__all__ = ['QueryResponse', 'GenericClient']


class QueryResponse(QueryResponseTable):
    hide_keys = ['url']

    def time_range(self):
        """
        Returns the time-span for which records are available.
        """
        if 'Start Time' in self.colnames and 'End Time' in self.colnames:
            return TimeRange(np.min(self['Start Time']), np.max(self['End Time']))

    def response_block_properties(self):
        """
        Returns a set of class attributes on all the response blocks.
        """
        s = {a if not a.startswith('_') else None for a in dir(self[0])}
        for resp in self[1:]:
            s = s.intersection({a if not a.startswith('_') else None for a in dir(resp)})

        s.remove(None)
        return s


class GenericClient(BaseClient):
    """
    Base class for simple web clients for the data retriever module. This class
    is mainly designed for downloading data from FTP and HTTP type data
    sources, although should in theory be general enough to get data from any
    web service.

    This class has two user facing methods
    `~sunpy.net.dataretriever.client.GenericClient.search` and
    `~sunpy.net.dataretriever.client.GenericClient.fetch` the former generates a
    set of results for files available through the service the client is
    querying and the latter downloads that data.

    Search uses two hooks as helper functions; these are
    :meth:`~sunpy.net.dataretriever.GenericClient.pre_search_hook` and
    :meth:`~sunpy.net.dataretriever.GenericClient.post_search_hook`.
    They help to translate the attrs for scraper before and after the search respectively.
    """
    baseurl = None
    # A regex string that can match all urls supported by the client.
    # A string which is used to extract the desired metadata from urls correctly,
    # using ``sunpy.extern.parse.parse``.
    pattern = None
    # Set of required 'attrs' for client to handle the query.
    required = {a.Time, a.Instrument}

    @classmethod
    def _get_match_dict(cls, *args, **kwargs):
        """
        Constructs a dictionary using the query and registered Attrs that represents
        all possible values of the extracted metadata for files that matches the query.
        The returned dictionary is used to validate the metadata of searched files
        in :func:`~sunpy.net.scraper.Scraper._extract_files_meta`.

        Parameters
        ----------
        \\*args: `tuple`
            `sunpy.net.attrs` objects representing the query.
        \\*\\*kwargs: `dict`
             Any extra keywords to refine the search.

        Returns
        -------
        matchdict: `dict`
            A dictionary having a `list` of all possible Attr values
            corresponding to an Attr.
        """
        regattrs_dict = cls.register_values()
        matchdict = {}
        for i in regattrs_dict.keys():
            attrname = i.__name__
            # only Attr values that are subclas of Simple Attr are stored as list in matchdict
            # since complex attrs like Range can't be compared with string matching.
            if issubclass(i, SimpleAttr):
                matchdict[attrname] = []
                for val, _ in regattrs_dict[i]:
                    matchdict[attrname].append(val)
        for elem in args:
            if isinstance(elem, a.Time):
                matchdict['Start Time'] = elem.start
                matchdict['End Time'] = elem.end
            elif hasattr(elem, 'value'):
                matchdict[elem.__class__.__name__] = [str(elem.value).lower()]
            elif isinstance(elem, a.Wavelength):
                matchdict['Wavelength'] = elem
            else:
                raise ValueError(
                    "GenericClient can not add {} to the rowdict dictionary to"
                    "pass to the Client.".format(elem.__class__.__name__))
        return matchdict

    @classmethod
    def pre_search_hook(cls, *args, **kwargs):
        """
        Helper function to return the baseurl, pattern and matchdict
        for the client required by :func:`~sunpy.net.dataretriever.GenericClient.search`
        before using the scraper.
        """
        matchdict = cls._get_match_dict(*args, **kwargs)
        return cls.baseurl, cls.pattern, matchdict

    @classmethod
    def _can_handle_query(cls, *query):
        """
        Method the
        `sunpy.net.fido_factory.UnifiedDownloaderFactory`
        class uses to dispatch queries to this Client.
        """
        regattrs_dict = cls.register_values()
        optional = {k for k in regattrs_dict.keys()} - cls.required
        if not cls.check_attr_types_in_query(query, cls.required, optional):
            return False
        for key in regattrs_dict:
            all_vals = [i[0].lower() for i in regattrs_dict[key]]
            for x in query:
                if isinstance(x, key) and issubclass(key, SimpleAttr) and str(x.value).lower() not in all_vals:
                    return False
        return True

    def post_search_hook(self, exdict, matchdict):
        """
        Helper function used after :func:`~sunpy.net.dataretriever.GenericClient.search`
        which makes the extracted metadata representable in a query response table.

        Parameters
        ----------
        exdict : `dict`
            Represents metadata extracted from files.
        matchdict : `dict`
            Contains attr values accessed from ``register_values()``
            and the search query itself.

        Returns
        -------
        rowdict: `~collections.OrderedDict`
            An Ordered Dictionary which is used by `QueryResponse`
            to show results.
        """
        rowdict = OrderedDict()
        tr = get_timerange_from_exdict(exdict)
        start = tr.start
        start.format = 'iso'
        end = tr.end
        end.format = 'iso'
        rowdict['Start Time'] = start
        rowdict['End Time'] = end
        for k in matchdict:
            if k not in ('Start Time', 'End Time', 'Wavelength'):
                if k == 'Physobs':
                    # not changing case for Phsyobs
                    rowdict[k] = matchdict[k][0]
                else:
                    rowdict[k] = matchdict[k][0].upper()
        for k in exdict:
            if k not in ['year', 'month', 'day', 'hour', 'minute', 'second']:
                rowdict[k] = exdict[k]
        return rowdict

    def _get_full_filenames(self, qres, filenames, path):
        """
        Returns full pathnames for each file in the result.

        Parameters
        ----------
        qres : `~sunpy.net.dataretriever.QueryResponse`
            Results to download.
        filenames : list
            List of base filenames (ex - "xyz.txt")
        path : str
            Path to download files to

        Returns
        -------
        List of full pathnames for each file (download_directory + filename)
        """
        # Create function to compute the filepath to download to if not set
        default_dir = Path(sunpy.config.get("downloads", "download_dir"))

        paths = []
        for i, filename in enumerate(filenames):
            fname = Path(filename)
            if path is None:
                fname = default_dir / '{file}'
            elif '{file}' not in str(path):
                fname = path / '{file}'
            else:
                fname = path

            temp_dict = qres[i].response_block_map
            temp_dict['file'] = str(filename)
            fname = fname.expanduser()
            fname = Path(str(fname).format(**temp_dict))

            paths.append(fname)

        return paths

    def search(self, *args, **kwargs):
        """
        Query this client for a list of results.

        Parameters
        ----------
        \\*args: `tuple`
            `sunpy.net.attrs` objects representing the query.
        \\*\\*kwargs: `dict`
             Any extra keywords to refine the search.

        Returns
        -------
        A `QueryResponse` instance containing the query result.
        """
        baseurl, pattern, matchdict = self.pre_search_hook(*args, **kwargs)
        scraper = Scraper(baseurl, regex=True)
        tr = TimeRange(matchdict['Start Time'], matchdict['End Time'])
        filesmeta = scraper._extract_files_meta(tr, extractor=pattern,
                                                matcher=matchdict)
        filesmeta = sorted(filesmeta, key=lambda k: k['url'])
        metalist = []
        for i in filesmeta:
            rowdict = self.post_search_hook(i, matchdict)
            metalist.append(rowdict)
        return QueryResponse(metalist, client=self)

    def fetch(self, qres, path=None, overwrite=False,
              progress=True, downloader=None, wait=True, **kwargs):
        """
        Download a set of results.

        Parameters
        ----------
        qres : `~sunpy.net.dataretriever.QueryResponse`
            Results to download.
        path : `str` or `pathlib.Path`, optional
            Path to the download directory, or file template including the
            ``{file}`` string which will be replaced with the filename.
        overwrite : `bool` or `str`, optional
            Determine how to handle downloading if a file already exists with the
            same name. If `False` the file download will be skipped and the path
            returned to the existing file, if `True` the file will be downloaded
            and the existing file will be overwritten, if ``'unique'`` the filename
            will be modified to be unique.
        progress : `bool`, optional
            If `True` show a progress bar showing how many of the total files
            have been downloaded. If `False`, no progress bar will be shown.
        downloader : `parfive.Downloader`, optional
            The download manager to use.
        wait : `bool`, optional
            If `False` ``downloader.download()`` will not be called. Only has
            any effect if ``downloader`` is not `None`.

        Returns
        -------
        `parfive.Results`
        """
        if path is not None:
            path = Path(path)

        if isinstance(qres, QueryResponseRow):
            qres = qres.as_table()

        urls = []
        if len(qres):
            urls = list(qres['url'])

        filenames = [url.split('/')[-1] for url in urls]

        paths = self._get_full_filenames(qres, filenames, path)

        dl_set = True
        if not downloader:
            dl_set = False
            downloader = Downloader(progress=progress, overwrite=overwrite)

        for url, filename in zip(urls, paths):
            downloader.enqueue_file(url, filename=filename)

        if dl_set and not wait:
            return

        return downloader.download()



from io import BytesIO

import asdf

# Remove this file when we depend on asdf 2.9+


def roundtrip_object(obj, version=None):
    """
    Add the specified object to an AsdfFile's tree, write the file to
    a buffer, then read it back in and return the deserialized object.

    Parameters
    ----------
    obj : object
        Object to serialize.
    version : str or None.
        ASDF Standard version.  If None, use the library's default version.

    Returns
    -------
    object
        The deserialized object.
    """
    buff = BytesIO()
    with asdf.AsdfFile(version=version) as af:
        af["obj"] = obj
        af.write_to(buff)

    buff.seek(0)
    with asdf.open(buff, lazy_load=False, copy_arrays=True) as af:
        return af["obj"]



from sunpy.net.attr import AttrAnd, AttrOr, AttrWalker
from sunpy.net.attrs import Time
from sunpy.net.cdaweb.attrs import Dataset

walker = AttrWalker()


@walker.add_creator(AttrOr)
def create_or(wlk, tree):
    results = []
    for sub in tree.attrs:
        results.append(wlk.create(sub))
    return results


@walker.add_creator(AttrAnd)
def create_and(wlk, tree):
    param_dict = {}
    wlk.apply(tree, param_dict)
    return [param_dict]


@walker.add_applier(AttrAnd)
def apply_and(wlk, and_attr, param_dict):
    for iattr in and_attr.attrs:
        wlk.apply(iattr, param_dict)


@walker.add_applier(Time)
def apply_timerange(wlk, time_attr, param_dict):
    param_dict['begin_time'] = time_attr.start
    param_dict['end_time'] = time_attr.end


@walker.add_applier(Dataset)
def _(wlk, attr, param_dict):
    param_dict['dataset'] = attr.value



import pathlib
import functools
from contextlib import contextmanager
from urllib.parse import urlparse

from sunpy.util.exceptions import warn_user
from sunpy.util.util import hash_file

__all__ = ['DataManager']


class DataManager:
    """
    This class provides a remote data manager for managing remote files.

    Parameters
    ----------
    cache: `sunpy.data.data_manager.cache.Cache`
        Cache object to be used by `~sunpy.data.data_manager.manager.DataManager`.
    """

    def __init__(self, cache):
        self._cache = cache

        self._file_cache = {}

        self._namespace = None
        self._skip_hash_check = False
        self._skip_file = {}

    def require(self, name, urls, sha_hash):
        """
        Decorator for informing the data manager about the requirement of
        a file by a function.

        Parameters
        ----------
        name : `str`
            The name to reference the file with.
        urls : `list` or `str`
            A list of urls to download the file from.
        sha_hash : `str`
            SHA-256 hash of file.
        """
        if isinstance(urls, str):
            urls = [urls]

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                self._namespace = self._get_module(func)
                replace = self._skip_file.get(name)
                if replace:
                    uri_parse = urlparse(replace['uri'])
                    if uri_parse.scheme in ("", "file"):
                        # If a relative file uri is specified (i.e.
                        # `file://sunpy/test`) this maintains compatibility
                        # with the original behaviour where this would be
                        # interpreted as `./sunpy/test` if no scheme is
                        # specified netloc will be '' by default.
                        file_path = uri_parse.netloc + uri_parse.path
                        file_hash = hash_file(file_path)
                    else:
                        file_path, file_hash, _ = self._cache._download_and_hash(
                            [replace['uri']], self._namespace
                        )
                    if replace['hash'] and file_hash != replace['hash']:
                        # if hash provided to replace function doesn't match the hash of the file
                        # raise error
                        raise ValueError(
                            "Hash provided to override_file does not match hash of the file.")
                elif self._skip_hash_check:
                    file_path = self._cache.download(urls, self._namespace, redownload=True)
                else:
                    details = self._cache.get_by_hash(sha_hash)
                    if not details:
                        # In case we are matching by hash and file does not exist
                        # That might mean the wrong hash is supplied to decorator
                        # We match by urls to make sure that is not the case
                        if self._cache_has_file(urls):
                            # If we can't find a file matching sha_hash, but the url is already
                            # in the database
                            raise ValueError(f"{urls} has already been downloaded, but no file "
                                             f"matching the hash {sha_hash} can be found.")
                        file_path = self._cache.download(urls, self._namespace)
                        file_hash = hash_file(file_path)
                        if file_hash != sha_hash:
                            # the hash of the file downloaded does not match provided hash
                            # this means the file has changed on the server.
                            # the function should be updated to use the new
                            # hash. Raise an error to notify.
                            raise RuntimeError(
                                f"Hash of local file ({file_hash}) does not match expected hash ({sha_hash}). "
                                "File may have changed on the remote server.")
                    else:
                        # This is to handle the case when the local file
                        # appears to be tampered/corrupted
                        if hash_file(details['file_path']) != details['file_hash']:
                            warn_user("Hashes do not match, the file will be redownloaded "
                                      "(could be be tampered/corrupted)")
                            file_path = self._cache.download(urls, self._namespace, redownload=True)
                            # Recheck the hash again, if this fails, we will exit.
                            if hash_file(file_path) != details['file_hash']:
                                raise RuntimeError("Redownloaded file also has the incorrect hash."
                                                   "The remote file on the server might have changed.")
                        else:
                            file_path = details['file_path']

                if name not in self._file_cache:
                    self._file_cache[name] = {}
                self._file_cache[name][self._namespace] = file_path
                result = func(*args, **kwargs)
                self._namespace = None
                return result
            return wrapper

        return decorator

    @contextmanager
    def override_file(self, name, uri, sha_hash=None):
        """
        Replaces the file by the name with the file provided by the url/path.

        Parameters
        ----------
        name : `str`
            Name of the file provided in the `require` decorator.
        uri : `str`
            URI of the file which replaces original file. Scheme should be one
            of ``http``, ``https``, ``ftp`` or ``file``. If no scheme is given
            the uri will be interpreted as a local path. i.e.
            ``file:///tmp/test`` and ``/tmp/test`` are the same.
        sha_hash : `str`, optional
            SHA256 hash of the file to compared to after downloading.
        """
        try:
            self._skip_file[name] = {
                'uri': uri,
                'hash': sha_hash,
            }
            yield
        finally:
            _ = self._skip_file.pop(name, None)

    @contextmanager
    def skip_hash_check(self):
        """
        Disables hash checking temporarily

        Examples
        --------
        >>> with remote_data_manager.skip_hash_check():  # doctest: +SKIP
        ...     myfunction()  # doctest: +SKIP
        """
        try:
            self._skip_hash_check = True
            yield
        finally:
            self._skip_hash_check = False

    def get(self, name):
        """
        Get the file by name.

        Parameters
        ----------
        name : `str`
            Name of the file given to the data manager, same as the one provided
            in `~sunpy.data.data_manager.manager.DataManager.require`.

        Returns
        -------
        `pathlib.Path`
            Path of the file.

        Raises
        ------
        `KeyError`
            If ``name`` is not in the cache.
        """
        return pathlib.Path(self._file_cache[name][self._namespace])

    def _cache_has_file(self, urls):
        for url in urls:
            if self._cache._get_by_url(url):
                return True
        return False

    def _get_module(self, func):
        """
        Returns the name of the module (appended with a dot) that a function belongs to.

        Parameters
        ----------
        func : function
            A function whose module is to be found.

        Returns
        -------
        A `str` that represents the module name appended with a dot.
        """
        module = func.__module__.lstrip('.').split('.')[0] + '.'
        if module == '__main__.':
            module = ''
        return module



import numpy as np
import pytest
from hypothesis import given, settings

import astropy.units as u
from astropy.coordinates import SkyCoord, SkyOffsetFrame
from astropy.tests.helper import assert_quantity_allclose

from sunpy.coordinates import NorthOffsetFrame
from sunpy.coordinates.tests.helpers import assert_longitude_allclose
from sunpy.coordinates.tests.strategies import latitudes, longitudes


def test_null():
    """
    test init of a frame where the origins are the same.
    """
    off = NorthOffsetFrame(north=SkyCoord(0*u.deg, 90*u.deg,
                                          frame='heliographic_stonyhurst'))
    assert isinstance(off, SkyOffsetFrame)
    assert off.origin.lat == 0*u.deg
    assert off.origin.lon == 0*u.deg


@given(lon=longitudes(), lat=latitudes())
@settings(deadline=1000)
def test_transform(lon, lat):
    """
    Test that the north pole in the new frame transforms back to the given
    north argument.
    """
    north = SkyCoord(lon=lon, lat=lat, frame='heliographic_stonyhurst')
    off = NorthOffsetFrame(north=north)
    t_north = SkyCoord(lon=0*u.deg, lat=90*u.deg, frame=off)
    t_north = t_north.transform_to('heliographic_stonyhurst')
    assert_longitude_allclose(north.lon, t_north.lon, atol=1e-6*u.deg)
    assert_quantity_allclose(north.lat, t_north.lat, atol=1e-6*u.deg)


def test_south_pole():
    s = SkyCoord(-10*u.deg, 0*u.deg, frame='heliographic_stonyhurst')
    off = NorthOffsetFrame(north=s)
    assert_longitude_allclose(off.origin.lon, 170*u.deg)
    assert_quantity_allclose(off.origin.lat, -90*u.deg)


def test_cartesian():
    # Test a north pole specified in Cartesian coordinates
    s = SkyCoord(1, 2, 3, unit=u.m,
                 representation_type='cartesian', frame='heliographic_stonyhurst')
    off = NorthOffsetFrame(north=s)
    assert_longitude_allclose(off.origin.lon, np.arctan2(s.y, s.x))
    assert_quantity_allclose(off.origin.lat, np.arctan2(s.z, np.sqrt(s.x**2 + s.y**2)) - 90*u.deg)


def test_error():
    with pytest.raises(TypeError):
        NorthOffsetFrame()



import os

import numpy as np
import pytest

import astropy.units as u

import sunpy.data.test
import sunpy.map
from sunpy.image.resample import reshape_image_to_4d_superpixel
from sunpy.util.exceptions import SunpyDeprecationWarning


@pytest.fixture
def aia171_test_map():
    testpath = sunpy.data.test.rootdir
    return sunpy.map.Map(os.path.join(testpath, 'aia_171_level1.fits'))


@pytest.fixture
def shape(aia171_test_map):
    return np.array(aia171_test_map.data.shape)


def resample_meta(aia171_test_map, dimensions, method, center, minusone):
    map_resampled = aia171_test_map.resample(dimensions, method=method)
    return tuple(map_resampled.data.shape)


def resample_method(aia171_test_map, method):
    for shape in [(64, 64), (256, 256)]:
        for center in [False, True]:
            for minusone in [False, True]:
                assert resample_meta(aia171_test_map, shape * u.pix, method, center, minusone) == shape


def test_resample_neighbor(aia171_test_map):
    with pytest.warns(SunpyDeprecationWarning,
                      match='Using "neighbor" as a method for resampling is deprecated.'):
        resample_method(aia171_test_map, 'neighbor')


def test_resample_nearest(aia171_test_map):
    resample_method(aia171_test_map, 'nearest')


def test_resample_linear(aia171_test_map):
    resample_method(aia171_test_map, 'linear')


def test_resample_spline(aia171_test_map):
    resample_method(aia171_test_map, 'spline')


def test_reshape(aia171_test_map, shape):

    def _n(a, b, c):
        return int(np.floor((a-b)/c))

    # Dimension divides the array shape exactly with no remainder
    im = reshape_image_to_4d_superpixel(aia171_test_map.data, (2, 2), (0, 0))
    assert im.shape == (shape[0]/2, 2, shape[1]/2, 2)
    # Dimension divides the array shape exactly with remainder
    im = reshape_image_to_4d_superpixel(aia171_test_map.data, (7, 5), (0, 0))
    assert im.shape == (int(shape[0]/7), 7, int(shape[1]/5), 5)
    # Dimension divides the array shape exactly with no remainder, and there is
    # an offset
    im = reshape_image_to_4d_superpixel(aia171_test_map.data, (2, 2), (1, 1))
    assert im.shape == (int(shape[0]/2) - 1, 2, int(shape[1]/2) - 1, 2)
    # Dimension divides the array shape exactly with remainder, and there is
    # an offset
    d = (9, 7)
    o = (1, 4)
    im = reshape_image_to_4d_superpixel(aia171_test_map.data, d, o)
    assert im.shape == (_n(shape[0], o[0], d[0]), d[0],
                        _n(shape[1], o[1], d[1]), d[1])
    im = reshape_image_to_4d_superpixel(aia171_test_map.data, d, o)
    assert im.shape == (_n(shape[0], o[0], d[0]), d[0],
                        _n(shape[1], o[1], d[1]), d[1])

    d = (9, 7)
    o = (5, 4)
    im = reshape_image_to_4d_superpixel(aia171_test_map.data, d, o)
    assert im.shape == (_n(shape[0], o[0], d[0]), d[0],
                        _n(shape[1], o[1], d[1]), d[1])

    d = (9, 7)
    o = (4, 4)
    im = reshape_image_to_4d_superpixel(aia171_test_map.data, d, o)
    assert im.shape == (_n(shape[0], o[0], d[0]), d[0],
                        _n(shape[1], o[1], d[1]), d[1])



"""
Datasource-specific classes

This is where datasource specific logic is implemented. Each mission should
have its own file with one or more classes defined. Typically, these classes
will be subclasses of the :mod`sunpy.map.Map` class.
"""

from ..map_factory import Map
from .hinode import *
from .iris import *
from .mlso import *
from .proba2 import *
from .psp import *
from .rhessi import *
from .sdo import *
from .soho import *
from .solo import *
from .source_type import *
from .stereo import *
from .suvi import *
from .trace import *
from .yohkoh import *



"""
==============================================
Overplotting the position of the Venus transit
==============================================

How to accurately plot the position of Venus as it transitted in front
of the Sun as observed by SDO/AIA.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord, solar_system_ephemeris

import sunpy.map
from sunpy.coordinates import get_body_heliographic_stonyhurst
from sunpy.net import Fido
from sunpy.net import attrs as a

###############################################################################
# Let's download an image of the Venus transit.

result = Fido.search(a.Time('2012/06/06 04:07:25', '2012/06/06 04:07:35'),
                     a.Instrument.aia,
                     a.Wavelength(1600*u.angstrom))
files = Fido.fetch(result)
aiamap = sunpy.map.Map(files[0])

###############################################################################
# For this example, we require high-precision ephemeris information. The built-in
# ephemeris provided by astropy is not accurate enough. This call requires ``jplephem``
# to be installed. This will also trigger a download of about ~10 MB.

solar_system_ephemeris.set('de432s')

###############################################################################
# Now we get the position of Venus and convert it into the SDO/AIA coordinates.
# The apparent position of Venus accounts for the time it takes for light to
# travel from Venus to SDO.

venus = get_body_heliographic_stonyhurst('venus', aiamap.date, observer=aiamap.observer_coordinate)
venus_hpc = venus.transform_to(aiamap.coordinate_frame)

###############################################################################
# Let's crop the image with Venus at its center.

fov = 200 * u.arcsec
bottom_left = SkyCoord(venus_hpc.Tx - fov/2, venus_hpc.Ty - fov/2, frame=aiamap.coordinate_frame)
smap = aiamap.submap(bottom_left, width=fov, height=fov)

###############################################################################
# Let's plot the results.

fig = plt.figure()
ax = fig.add_subplot(projection=smap)
smap.plot(axes=ax)
smap.draw_limb(axes=ax)
ax.grid(False)
ax.plot_coord(venus_hpc, 'x', color='deepskyblue', label='Venus')
ax.legend()

plt.show()



import cdflib
import pandas as pd
from cdflib.epochs import CDFepoch

import astropy.units as u

from sunpy import log
from sunpy.timeseries import GenericTimeSeries
from sunpy.util.exceptions import warn_user

__all__ = ['read_cdf']


def read_cdf(fname):
    """
    Read a CDF file that follows the ISTP/IACG guidelines.

    Parameters
    ----------
    fname : path-like
        Location of single CDF file to read.

    Returns
    -------
    list[GenericTimeSeries]
        A list of time series objects, one for each unique time index within
        the CDF file.

    References
    ----------
    Space Physics Guidelines for CDF https://spdf.gsfc.nasa.gov/sp_use_of_cdf.html
    """
    cdf = cdflib.CDF(str(fname))

    # Extract the time varying variables
    cdf_info = cdf.cdf_info()
    meta = cdf.globalattsget()
    all_var_keys = cdf_info['rVariables'] + cdf_info['zVariables']
    var_attrs = {key: cdf.varattsget(key) for key in all_var_keys}
    # Get keys that depend on time
    var_keys = [var for var in var_attrs if 'DEPEND_0' in var_attrs[var]]

    # Get unique time index keys
    time_index_keys = sorted(set([var_attrs[var]['DEPEND_0'] for var in var_keys]))

    all_ts = []
    # For each time index, construct a GenericTimeSeries
    for index_key in time_index_keys:
        try:
            index = cdf.varget(index_key)
        except ValueError:
            # Empty index for cdflib >= 0.3.20
            continue
        if index is None:
            # Empty index for cdflib <0.3.20
            continue
        # TODO: use to_astropy_time() instead here when we drop pandas in timeseries
        index = CDFepoch.to_datetime(index)
        df = pd.DataFrame(index=pd.DatetimeIndex(name=index_key, data=index))
        units = {}

        for var_key in sorted(var_keys):
            attrs = var_attrs[var_key]
            if attrs['DEPEND_0'] != index_key:
                continue

            # Get data
            if cdf.varinq(var_key)['Last_Rec'] == -1:
                log.debug(f'Skipping {var_key} in {fname} as it has zero elements')
                continue

            data = cdf.varget(var_key)
            # Get units
            if 'UNITS' in attrs:
                unit_str = attrs['UNITS']
                try:
                    unit = u.Unit(unit_str)
                except ValueError:
                    if unit_str in _known_units:
                        unit = _known_units[unit_str]
                    else:
                        warn_user(f'astropy did not recognize units of "{unit_str}". '
                                  'Assigning dimensionless units. '
                                  'If you think this unit should not be dimensionless, '
                                  'please raise an issue at https://github.com/sunpy/sunpy/issues')
                        unit = u.dimensionless_unscaled
            else:
                warn_user(f'No units provided for variable "{var_key}". '
                          'Assigning dimensionless units.')
                unit = u.dimensionless_unscaled

            if data.ndim > 2:
                # Skip data with dimensions >= 3 and give user warning
                warn_user(f'The variable "{var_key}" has been skipped because it has more than 2 dimensions, which is unsupported.')
            elif data.ndim == 2:
                # Multiple columns, give each column a unique label
                for i, col in enumerate(data.T):
                    df[var_key + f'_{i}'] = col
                    units[var_key + f'_{i}'] = unit
            else:
                # Single column
                df[var_key] = data
                units[var_key] = unit

        all_ts.append(GenericTimeSeries(data=df, units=units, meta=meta))

    if not len(all_ts):
        log.debug(f'No data found in file {fname}')
    return all_ts


_known_units = {'ratio': u.dimensionless_unscaled,
                'NOTEXIST': u.dimensionless_unscaled,
                'Unitless': u.dimensionless_unscaled,
                'unitless': u.dimensionless_unscaled,
                'Quality_Flag': u.dimensionless_unscaled,
                'None': u.dimensionless_unscaled,
                'none': u.dimensionless_unscaled,
                ' none': u.dimensionless_unscaled,
                'counts': u.dimensionless_unscaled,
                'cnts': u.dimensionless_unscaled,

                'microW m^-2': u.mW * u.m**-2,

                'years': u.yr,
                'days': u.d,

                '#/cc': u.cm**-3,
                '#/cm^3': u.cm**-3,
                'cm^{-3}': u.cm**-3,
                'particles cm^-3': u.cm**-3,
                'n/cc (from moments)': u.cm**-3,
                'n/cc (from fits)': u.cm**-3,
                'Per cc': u.cm**-3,
                '#/cm3': u.cm**-3,
                'n/cc': u.cm**-3,

                'km/sec': u.km / u.s,
                'km/sec (from fits)': u.km / u.s,
                'km/sec (from moments)': u.km / u.s,
                'Km/s': u.km / u.s,

                'Volts': u.V,

                'earth radii': u.earthRad,
                'Re': u.earthRad,
                'Earth Radii': u.earthRad,
                'Re (1min)': u.earthRad,
                'Re (1hr)': u.earthRad,

                'Degrees': u.deg,
                'degrees': u.deg,
                'Deg': u.deg,
                'deg (from fits)': u.deg,
                'deg (from moments)': u.deg,
                'deg (>200)': u.deg,

                'Deg K': u.K,
                'deg_K': u.K,
                '#/{cc*(cm/s)^3}': (u.cm**3 * (u.cm / u.s)**3)**-1,
                'sec': u.s,
                'Samples/s': 1 / u.s,

                'seconds': u.s,
                'nT GSE': u.nT,
                'nT GSM': u.nT,
                'nT DSL': u.nT,
                'nT SSL': u.nT,
                'nT (1min)': u.nT,
                'nT (3sec)': u.nT,
                'nT (1hr)': u.nT,
                'nT (>200)': u.nT,

                'msec': u.ms,
                'milliseconds': u.ms,

                '#/cm2-ster-eV-sec': 1 / (u.cm**2 * u.sr * u.eV * u.s),
                '#/(cm^2*s*sr*MeV/nuc)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                '#/(cm^2*s*sr*Mev/nuc)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                '#/(cm^2*s*sr*Mev/nucleon)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                '#/(cm2-steradian-second-MeV/nucleon) ': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                '1/(cm2 Sr sec MeV/nucleon)': 1 / (u.cm**2 * u.sr * u.s * u.MeV),
                '1/(cm**2-s-sr-MeV)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                '1/(cm**2-s-sr-MeV/nuc.)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                '1/(cm^2 sec ster MeV)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                'cnts/sec/sr/cm^2/MeV': 1 / (u.cm**2 * u.s * u.sr * u.MeV),

                'particles / (s cm^2 sr MeV)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                'particles / (s cm^2 sr MeV/n)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),
                'particles/(s cm2 sr MeV/n)': 1 / (u.cm**2 * u.s * u.sr * u.MeV),

                '1/(cm**2-s-sr)': 1 / (u.cm**2 * u.s * u.sr),
                '1/(SQcm-ster-s)': 1 / (u.cm**2 * u.s * u.sr),
                '1/(SQcm-ster-s)..': 1 / (u.cm**2 * u.s * u.sr),

                'Counts/256sec': 1 / (256 * u.s),
                'Counts/hour': 1 / u.hr,
                'counts / s': 1/u.s,
                'cnts/sec': 1/u.s,
                }



from packaging.requirements import Requirement

from sunpy.util.sysinfo import (
    find_dependencies,
    format_requirement_string,
    missing_dependencies_by_extra,
    resolve_requirement_versions,
    system_info,
)


def test_find_dependencies():
    missing, installed = find_dependencies()
    assert missing == {}
    assert sorted(list(installed.keys())) == sorted(["astropy", "numpy", "packaging", "parfive"])

    missing, installed = find_dependencies(package="sunpy", extras=["required", "all"])
    assert missing == {}
    assert sorted(list(installed.keys())) == sorted(['asdf',
                                                     'asdf-astropy',
                                                     'astropy',
                                                     'numpy',
                                                     'parfive',
                                                     'packaging',
                                                     'dask',
                                                     'sqlalchemy',
                                                     'scikit-image',
                                                     'scipy',
                                                     'glymur',
                                                     'matplotlib',
                                                     'mpl-animators',
                                                     'reproject',
                                                     'beautifulsoup4',
                                                     'drms',
                                                     'python-dateutil',
                                                     'tqdm',
                                                     'zeep',
                                                     'cdflib',
                                                     'h5netcdf',
                                                     'h5py',
                                                     'pandas'])


def test_missing_dependencies_by_extra():
    missing = missing_dependencies_by_extra()
    assert sorted(list(missing.keys())) == sorted(['all',
                                                   'asdf',
                                                   'required',
                                                   'dask',
                                                   'database',
                                                   'dev',
                                                   'docs',
                                                   'image',
                                                   'jpeg2000',
                                                   'map',
                                                   'net',
                                                   'tests',
                                                   'timeseries',
                                                   'visualization'])
    missing = missing_dependencies_by_extra(exclude_extras=["all"])
    assert sorted(list(missing.keys())) == sorted(['asdf', 'required', 'dask', 'database', 'dev', 'docs',
                                                   'image', 'jpeg2000', 'map', 'net', 'tests', 'timeseries',
                                                   'visualization'])


def test_resolve_requirement_versions():
    package1 = Requirement('test-package[ext1]>=1.1.1; extra == "group1"')
    package2 = Requirement('test-package[ext2]<=2.0.0; extra == "group2"')
    assert str(resolve_requirement_versions([package1, package2])) == str(Requirement(
        'test-package[ext1,ext2]<=2.0.0,>=1.1.1; extra == "group1" or extra == "group2"'))

    package3 = Requirement('test-package==1.1.0; extra == "group3"')
    package4 = Requirement('test-package==1.1.0; extra == "group4"')
    assert str(resolve_requirement_versions([package3, package4])) == str(
        Requirement('test-package==1.1.0; extra == "group3" or extra == "group4"'))

    package5 = Requirement('test-package; extra == "group5"')
    package6 = Requirement('test-package[ext3]@https://foo.com')
    assert str(resolve_requirement_versions([package5, package6])) == str(
        Requirement('test-package[ext3]@ https://foo.com ; extra == "group5"'))


def test_format_requirement_string():
    package1 = Requirement('test-package[ext1]>=1.1.1; extra == "group1"')
    assert format_requirement_string(package1) == 'Missing test-package[ext1]>=1.1.1; extra == "group1"'

    package2 = Requirement('test-package>=1.1.1; extra == "group1" or extra == "group2" or extra == "group3"')
    assert format_requirement_string(
        package2) == 'Missing test-package>=1.1.1; extra == "group1" or "group2" or "group3"'

    package3 = Requirement('test-package>=1.1.1')
    assert format_requirement_string(package3) == 'Missing test-package>=1.1.1'


def test_system_info(capsys):
    system_info()
    captured = capsys.readouterr()
    assert "\nsunpy Installation Information\n" in captured.out



# Author: Simon Liedtke <liedtke.simon@googlemail.com>
#
# This module was developed with funding provided by
# the Google Summer of Code (2013).
import os
import fnmatch
from datetime import datetime

import numpy as np
from sqlalchemy import Boolean, Column, DateTime, Float, ForeignKey, Integer, String, Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

import astropy.table
import astropy.units as u
from astropy.time import Time
from astropy.units import equivalencies

import sunpy
import sunpy.net.vso.legacy_response
from sunpy import config
from sunpy.io import _fits
from sunpy.io import file_tools as sunpy_filetools
from sunpy.io.header import FileHeader
from sunpy.time import parse_time
from sunpy.util.types import DatabaseEntryType

TIME_FORMAT = config.get("general", "time_format")

DEFAULT_HEADER = FileHeader([('SIMPLE', True),
                             ('BITPIX', 8),
                             ('NAXIS', 0),
                             ('EXTEND', True),
                             ('COMMENT', ''),
                             ('HISTORY', ''),
                             ('KEYCOMMENTS',
                              {'SIMPLE': 'conforms to FITS standard',
                               'BITPIX': 'array data type',
                               'NAXIS': 'number of array dimensions'}),
                             ('WAVEUNIT', None)])


__all__ = [
    'WaveunitNotFoundError', 'WaveunitNotConvertibleError', 'JSONDump',
    'FitsHeaderEntry', 'FitsKeyComment', 'Tag', 'DatabaseEntry',
    'entries_from_query_result', 'entries_from_file', 'entries_from_dir',
    'display_entries']

Base = declarative_base()

# required for the many-to-many relation on tags:entries
association_table = Table('association', Base.metadata,
                          Column('tag_name', String, ForeignKey('tags.name')),
                          Column('entry_id', Integer, ForeignKey('data.id'))
                          )


class WaveunitNotFoundError(Exception):
    """This exception is raised if a wavelength unit cannot be found in a FITS
    header or in a VSO query result block.

    """

    def __init__(self, obj):
        self.obj = obj

    def __str__(self):
        return f'the wavelength unit cannot be found in {self.obj}' + \
               ' and default_waveunit not specified when opening the database'


class WaveunitNotConvertibleError(Exception):
    """This exception is raised if a wavelength cannot be converted to an
    astropy.units.Unit instance.

    """

    def __init__(self, waveunit):
        self.waveunit = waveunit

    def __str__(self):
        return (
            'the waveunit {!r} cannot be converted to an '
            'astropy.units.Unit instance'.format(self.waveunit))


class JSONDump(Base):
    __tablename__ = 'jsondumps'

    dump = Column(String, nullable=False, primary_key=True)

    def __init__(self, dump):
        self.dump = dump

    def __eq__(self, other):
        return self.dump == other.dump

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return self.dump

    def __repr__(self):
        return f'{self.__class__.__name__}(dump={self.dump!r})'


class FitsHeaderEntry(Base):
    __tablename__ = 'fitsheaderentries'

    dbentry_id = Column(Integer, ForeignKey('data.id'))
    id = Column(Integer, primary_key=True)
    key = Column(String, nullable=False)
    value = Column(String)

    def __init__(self, key, value):
        self.key = key
        self.value = value

    def __eq__(self, other):
        return (
            (self.id == other.id or self.id is None or other.id is None) and
            self.key == other.key and
            self.value == other.value)

    def __hash__(self):
        return super().__hash__()

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return '{}(id={}, key={!r}, value={!r})'.format(
            self.__class__.__name__, self.id, self.key, self.value)


class FitsKeyComment(Base):
    __tablename__ = 'fitskeycomments'

    dbentry_id = Column(Integer, ForeignKey('data.id'))
    id = Column(Integer, primary_key=True)
    key = Column(String, nullable=False)
    value = Column(String)

    def __init__(self, key, value):
        self.key = key
        self.value = value

    def __eq__(self, other):
        return (
            (self.id == other.id or self.id is None or other.id is None) and
            self.key == other.key and
            self.value == other.value)

    def __lt__(self, other):
        return (f'{self.key}, {self.value}' <
                f'{other.key}, {other.value}')

    def __hash__(self):
        return super().__hash__()

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return '{}(id={}, key={!r}, value={!r})>'.format(
            self.__class__.__name__, self.id, self.key, self.value)


class Tag(Base):
    __tablename__ = 'tags'

    name = Column(String, nullable=False, primary_key=True)

    def __init__(self, name):
        self.name = name

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return super().__hash__()

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'{self.__class__.__name__}(name={self.name!r})'


class DatabaseEntry(DatabaseEntryType, Base):
    """
    This represents the main table of the database
    and each instance represents one record that *can* be saved in the
    database.

    Parameters
    ----------
    id : `int`
        A unique ID number. By default it is None, but automatically set to the
        maximum number plus one when this entry is added to the database.
    source : str
        The source is the name of an observatory or the name of a network of
        observatories.
    provider : str
        The name of the server which provides the retrieved data.
    physobs : str
        A physical observable identifier used by VSO.
    fileid : str
        The file ID is a string defined by the data provider that should point
        to a specific data product. The association of fileid to the specific
        data may change sometime, if the fileid always points to the latest
        calibrated data.
    observation_time_start : datetime.datetime
        The date and time when the observation of the data started.
    observation_time_end : datetime.datetime
        The date and time when the observation of the data ended.
    instrument : str
        The instrument which was used to observe the data.
    size : float
        The size of the data in kilobytes.
    wavemin : float
        The value of the measured wave length.
    wavemax : float
        This is the same value as ``wavemin``. The value is stored twice,
        because each `sunpy.net.dataretriever.client.QueryResponse` which is used by
        the vso package contains both these values.
    hdu_index : `int`
        This value provides a list of all available HDUs and in what
        files they are located.
    path : str
        A local file path where the according FITS file is saved.
    download_time : datetime.datetime
        The date and time when the files connected to a query have been
        downloaded. Note: this is not the date and time when this entry has
        been added to a database!
    starred : bool
        Entries can be starred to mark them. By default, this value is False.
    fits_header_entries : list
        A list of ``FitsHeaderEntry`` instances.
    tags : list
        A list of ``Tag`` instances. Use `sunpy.database.Database.tag` to
        add a new tag or multiple tags to a specific entry.
    """
    __tablename__ = 'data'

    # FIXME: primary key is data provider + file ID + download_time!
    id = Column(Integer, primary_key=True)
    source = Column(String)
    provider = Column(String)
    physobs = Column(String)
    fileid = Column(String)
    observation_time_start = Column(DateTime)
    observation_time_end = Column(DateTime)
    instrument = Column(String)
    size = Column(Float)
    wavemin = Column(Float)
    wavemax = Column(Float)
    hdu_index = Column(Integer)
    path = Column(String)
    download_time = Column(DateTime)
    starred = Column(Boolean, default=False)
    fits_header_entries = relationship('FitsHeaderEntry')
    fits_key_comments = relationship('FitsKeyComment')
    tags = relationship('Tag', secondary=association_table, backref='data')

    @classmethod
    def _from_query_result_block(cls, qr_block, default_waveunit=None):
        """Make a new :class:`DatabaseEntry` instance from a VSO query result
        block. The values of :attr:`wavemin` and :attr:`wavemax` are converted
        to nm (nanometres).

        Parameters
        ----------
        qr_block : `sunpy.net.dataretriever.client.QueryResponse`
            A query result block is usually not created directly; instead,
            one gets instances of `sunpy.net.dataretriever.client.QueryResponse` by
            iterating over a VSO query result.
        default_waveunit : str, optional
            The wavelength unit that is used if it cannot be found in the
            ``qr_block``.

        Examples
        --------
        >>> from sunpy.net import vso, attrs as a
        >>> from sunpy.database.tables import DatabaseEntry
        >>> client = vso.VSOClient()  # doctest: +REMOTE_DATA
        >>> qr = client.search(
        ...     a.Time('2001/1/1', '2001/1/2'),
        ...     a.Instrument.eit,
        ...     response_format="legacy")  # doctest: +REMOTE_DATA
        >>> entry = DatabaseEntry._from_query_result_block(qr.blocks[0])  # doctest: +REMOTE_DATA
        >>> entry.source  # doctest: +REMOTE_DATA
        'SOHO'
        >>> entry.provider  # doctest: +REMOTE_DATA
        'SDAC'
        >>> entry.physobs  # doctest: +REMOTE_DATA
        'intensity'
        >>> entry.fileid  # doctest: +REMOTE_DATA
        '/archive/soho/private/data/processed/eit/lz/2001/01/efz20010101.000042'
        >>> entry.observation_time_start, entry.observation_time_end  # doctest: +REMOTE_DATA
        (datetime.datetime(2001, 1, 1, 0, 0, 42), datetime.datetime(2001, 1, 1, 0, 0, 54))
        >>> entry.instrument  # doctest: +REMOTE_DATA
        'EIT'
        >>> entry.size  # doctest: +REMOTE_DATA
        2059.0
        >>> entry.wavemin, entry.wavemax  # doctest: +REMOTE_DATA
        (19.5, 19.5)

        """
        time_start = datetime.strptime(qr_block.time.start, '%Y%m%d%H%M%S')
        if not qr_block.time.end:
            qr_block.time.end = qr_block.time.start
        time_end = datetime.strptime(qr_block.time.end, '%Y%m%d%H%M%S')
        wave = qr_block.wave
        unit = None
        if wave.waveunit is None:
            if default_waveunit is not None:
                unit = u.Unit(default_waveunit)
        else:
            # some query response blocks store the unit "kev",
            # but Astropy only understands "keV". See issue #766.
            waveunit = wave.waveunit
            if waveunit == "kev":
                waveunit = "keV"
            unit = u.Unit(waveunit)
        if wave.wavemin is None:
            wavemin = None
        else:
            if unit is None:
                raise WaveunitNotFoundError(qr_block)
            wavemin = unit.to(u.nm, float(wave.wavemin),
                              equivalencies.spectral())
        if wave.wavemax is None:
            wavemax = None
        else:
            if unit is None:
                raise WaveunitNotFoundError(qr_block)
            wavemax = unit.to(u.nm, float(wave.wavemax),
                              equivalencies.spectral())
        source = getattr(qr_block, 'source', None)
        provider = getattr(qr_block, 'provider', None)
        fileid = getattr(qr_block, 'fileid', None)
        instrument = getattr(qr_block, 'instrument', None)
        size = getattr(qr_block, 'size', -1)
        physobs = getattr(qr_block, 'physobs', None)
        if physobs is not None:
            physobs = str(physobs)
        return cls(
            source=source, provider=provider, physobs=physobs, fileid=fileid,
            observation_time_start=time_start, observation_time_end=time_end,
            instrument=instrument, size=size,
            wavemin=wavemin, wavemax=wavemax)

    @classmethod
    def _from_fido_search_result_block(cls, sr_block, default_waveunit=None):
        """
        Make a new :class:`DatabaseEntry` instance from a Fido search
        result block.

        Parameters
        ----------
        sr_block : `sunpy.net.dataretriever.client.QueryResponse`
            A query result block is usually not created directly; instead,
            one gets instances of
            ``sunpy.net.dataretriever.client.QueryResponse`` by iterating
            over each element of a Fido search result.
        default_waveunit : `str`, optional
            The wavelength unit that is used if it cannot be found in the
            `sr_block`.
        """
        # All attributes of DatabaseEntry that are not in QueryResponse
        # are set as None for now.
        source = sr_block.get('Source')
        provider = sr_block.get('Provider')
        physobs = sr_block.get('Physobs')
        if physobs is not None:
            physobs = str(physobs)
        instrument = sr_block.get('Instrument')
        time_start = sr_block.get('Start Time')
        if time_start is not None:
            time_start = time_start.datetime
        time_end = sr_block.get('End Time')
        if time_end is not None:
            time_end = time_end.datetime

        wavelengths = sr_block.get('Wavelength', np.nan * u.nm)
        if wavelengths is None:
            wavelengths = np.nan * u.nm
        if isinstance(wavelengths, u.Quantity):
            if wavelengths.isscalar:
                wavemin = wavemax = wavelengths.to_value(u.nm, equivalencies=u.spectral())
            else:
                wavemin, wavemax = wavelengths.to_value(u.nm, equivalencies=u.spectral())
        else:
            raise TypeError("Expected Wavelength in the Fido response to be None or a Quantity")

        fileid = sr_block.get('url', sr_block.get('fileid'))
        size = None
        return cls(
            source=source, provider=provider, physobs=physobs, fileid=fileid,
            observation_time_start=time_start, observation_time_end=time_end,
            instrument=instrument, size=size,
            wavemin=wavemin, wavemax=wavemax)

    def __eq__(self, other):

        if self.wavemin is None and other.wavemin is None:
            wavemins_equal = True
        elif not all([self.wavemin, other.wavemin]):
            # This means one is None and the other isnt
            wavemins_equal = False
        else:
            wavemins_equal = np.allclose([self.wavemin], [other.wavemin], equal_nan=True)

        if self.wavemax is None and other.wavemax is None:
            wavemaxs_equal = True
        elif not all([self.wavemax, other.wavemax]):
            # This means one is None and the other isnt
            wavemaxs_equal = False
        else:
            wavemaxs_equal = np.allclose([self.wavemax], [other.wavemax], equal_nan=True)

        return (
            (self.id == other.id or self.id is None or other.id is None) and
            self.source == other.source and
            self.provider == other.provider and
            self.physobs == other.physobs and
            self.fileid == other.fileid and
            self.observation_time_start == other.observation_time_start and
            self.observation_time_end == other.observation_time_end and
            self.instrument == other.instrument and
            self.size == other.size and
            wavemins_equal and
            wavemaxs_equal and
            self.path == other.path and
            self.download_time == other.download_time and
            bool(self.starred) == bool(other.starred) and
            self.fits_header_entries == other.fits_header_entries and
            self.tags == other.tags)

    def _compare_attributes(self, other, attribute_list):
        """
        Compare a given list of attributes of two :class:`DatabaseEntry`
        instances and return True if all of them match.

        Parameters
        ----------
        other : :class:`DatabaseEntry` instance
        attribute_list : `list`
            The list of attributes that will be compared in both instances,
            self and other.

        """
        if len(attribute_list) == 0:
            raise TypeError('At least one attribute required')
        for attribute in attribute_list:
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True

    def __hash__(self):
        return super().__hash__()

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        attrs = [
            'id', 'source', 'provider', 'physobs', 'fileid',
            'observation_time_start', 'observation_time_end', 'instrument',
            'size', 'wavemin', 'wavemax', 'path', 'download_time', 'starred',
            'fits_header_entries', 'tags']
        ret = f'{self.__class__.__name__}('
        for attr in attrs:
            value = getattr(self, attr, None)
            if value:
                ret += f'{attr}={value!r}, '
        ret = ret.rstrip(', ')
        ret += ')'
        return ret


def entries_from_query_result(qr, default_waveunit=None):
    """
    Use a query response returned from :meth:`sunpy.net.vso.VSOClient.search`
    to generate instances of :class:`DatabaseEntry`. Return an iterator over
    those instances.

    Parameters
    ----------
    qr : `sunpy.net.vso.VSOQueryResponseTable`
        The query response from which to build the database entries.
    default_waveunit : `str`, optional
        The wavelength unit that is used if it cannot be found in the
        ``qr_block``.

    Examples
    --------
    >>> from sunpy.net import vso, attrs as a
    >>> from sunpy.database.tables import entries_from_query_result
    >>> client = vso.VSOClient()  # doctest: +REMOTE_DATA
    >>> qr = client.search(
    ...     a.Time('2001/1/1', '2001/1/2'),
    ...     a.Instrument.eit,
    ...     response_format="legacy")  # doctest: +REMOTE_DATA
    >>> entries = entries_from_query_result(qr)  # doctest: +REMOTE_DATA
    >>> entry = next(entries)  # doctest: +REMOTE_DATA
    >>> entry.source  # doctest: +REMOTE_DATA
    'SOHO'
    >>> entry.provider  # doctest: +REMOTE_DATA
    'SDAC'
    >>> entry.physobs  # doctest: +REMOTE_DATA
    'intensity'
    >>> entry.fileid  # doctest: +REMOTE_DATA
    '/archive/soho/private/data/processed/eit/lz/2001/01/efz20010101.000042'
    >>> entry.observation_time_start, entry.observation_time_end  # doctest: +REMOTE_DATA
    (datetime.datetime(2001, 1, 1, 0, 0, 42), datetime.datetime(2001, 1, 1, 0, 0, 54))
    >>> entry.instrument  # doctest: +REMOTE_DATA
    'EIT'
    >>> entry.size  # doctest: +REMOTE_DATA
    2059.0
    >>> entry.wavemin, entry.wavemax  # doctest: +REMOTE_DATA
    (19.5, 19.5)

    """
    for block in qr:
        yield DatabaseEntry._from_query_result_block(block, default_waveunit)


def entries_from_fido_search_result(sr, default_waveunit=None):
    """
    Use a `sunpy.net.dataretriever.fido_factory.UnifiedResponse`
    object returned from
    :meth:`sunpy.net.dataretriever.fido_factory.UnifiedDownloaderFactory.search`
    to generate instances of :class:`DatabaseEntry`. Return an iterator
    over those instances.

    Parameters
    ----------
    sr : `sunpy.net.dataretriever.fido_factory.UnifiedResponse`
        A UnifiedResponse object that is used to store responses from the
        unified downloader. This is returned by the ``search`` method of a
        :class:`sunpy.net.dataretriever.fido_factory.UnifiedDownloaderFactory`
        object.
    default_waveunit : `str`, optional
        The wavelength unit that is used if it cannot be found in the Query
        Response block.

    Examples
    --------
    >>> from sunpy.net import Fido, attrs
    >>> from sunpy.database.tables import entries_from_fido_search_result
    >>> sr = Fido.search(attrs.Time("2012/1/1", "2012/1/2"),
    ...     attrs.Instrument('lyra')) # doctest: +REMOTE_DATA
    >>> entries = entries_from_fido_search_result(sr) # doctest: +REMOTE_DATA
    >>> entry = next(entries) # doctest: +REMOTE_DATA
    >>> entry.source # doctest: +REMOTE_DATA
    'PROBA2'
    >>> entry.provider # doctest: +REMOTE_DATA
    'ESA'
    >>> entry.physobs # doctest: +REMOTE_DATA
    'irradiance'
    >>> entry.fileid # doctest: +REMOTE_DATA
    'http://proba2.oma.be/lyra/data/bsd/2012/01/01/lyra_20120101-000000_lev2_std.fits'
    >>> entry.observation_time_start, entry.observation_time_end # doctest: +REMOTE_DATA
    (datetime.datetime(2012, 1, 1, 0, 0),  datetime.datetime(2012, 1, 1, 23, 59, 59, 999000))
    >>> entry.instrument # doctest: +REMOTE_DATA
    'LYRA'

    """
    for entry in sr:
        if isinstance(entry, sunpy.net.vso.legacy_response.QueryResponse):
            # This is because Fido can search the VSO. It
            # returns a VSO QueryResponse.
            for block in entry:
                yield DatabaseEntry._from_query_result_block(block, default_waveunit)
        elif isinstance(entry, sunpy.net.jsoc.jsoc.JSOCResponse):
            # Adding JSOC results to the DB not supported for now
            raise ValueError("Cannot add JSOC results to database")
        else:
            for block in entry:
                yield DatabaseEntry._from_fido_search_result_block(block, default_waveunit)


def entries_from_file(file, default_waveunit=None,
                      time_string_parse_format=''):
    # Note: time_string_parse_format='' so that None won't be passed to Time.strptime
    # (which would make strptime freak out, if I remember correctly).
    """Use the headers of a FITS file to generate an iterator of
    :class:`sunpy.database.tables.DatabaseEntry` instances. Gathered
    information will be saved in the attribute ``fits_header_entries``. If the
    key INSTRUME, WAVELNTH or DATE-OBS / DATE_OBS is available, the attribute
    ``instrument``, ``wavemin`` and ``wavemax`` or ``observation_time_start`` is set,
    respectively. If the wavelength unit can be read, the values of ``wavemin``
    and ``wavemax`` are converted to nm (nanometres). The value of the ``file``
    parameter is used to set the attribute ``path`` of each generated database
    entry.

    Parameters
    ----------
    file : str, file object
        Either a path pointing to a FITS file or a an opened file-like object.
        If an opened file object, its mode must be one of the following rb,
        rb+, or ab+.
    default_waveunit : str, optional
        The wavelength unit that is used for a header if it cannot be found.
    time_string_parse_format : str, optional
        Fallback timestamp format which will be passed to
        `~astropy.time.Time.strptime` if `sunpy.time.parse_time` is unable to
        automatically read the ``date-obs`` metadata.

    Raises
    ------
    sunpy.database.tables.WaveunitNotFoundError
        If ``default_waveunit`` is not given and the wavelength unit cannot
        be found in one of the FITS headers
    sunpy.database.tables.WaveunitNotConvertibleError
        If a wavelength unit could be found but cannot be used to create an
        instance of the type ``astropy.units.Unit``. This can be the case
        for example if a FITS header has the key ``WAVEUNIT`` with the value
        ``nonsense``.

    Examples
    --------
    >>> from sunpy.database.tables import entries_from_file
    >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
    >>> entries = list(entries_from_file(sunpy.data.sample.SWAP_LEVEL1_IMAGE))  # doctest: +REMOTE_DATA
    >>> len(entries)  # doctest: +REMOTE_DATA
    1
    >>> entry = entries.pop()  # doctest: +REMOTE_DATA
    >>> entry.instrument  # doctest: +REMOTE_DATA
    'SWAP'
    >>> entry.observation_time_start, entry.observation_time_end  # doctest: +REMOTE_DATA
    (datetime.datetime(2011, 6, 7, 6, 33, 29, 759000), None)
    >>> entry.wavemin, entry.wavemax  # doctest: +REMOTE_DATA
    (17.400000000000002, 17.400000000000002)
    >>> len(entry.fits_header_entries)  # doctest: +REMOTE_DATA
    110

    """
    headers = _fits.get_header(file)

    # This just checks for blank default headers
    # due to compression.
    for header in headers:
        if header == DEFAULT_HEADER:
            headers.remove(header)

    if isinstance(file, str):
        filename = file
    else:
        filename = getattr(file, 'name', None)
    for header in headers:
        entry = DatabaseEntry(path=filename)
        for key, value in header.items():
            # Yes, it is possible to have an empty key in a FITS file.
            # Example: sunpy.data.sample.EIT_195_IMAGE
            # Don't ask me why this could be a good idea.
            if key == '':
                value = str(value)
            elif key == 'KEYCOMMENTS':
                for k, v in value.items():
                    entry.fits_key_comments.append(FitsKeyComment(k, v))
                continue
            entry.fits_header_entries.append(FitsHeaderEntry(key, value))
        waveunit = _fits.extract_waveunit(header)
        entry.hdu_index = headers.index(header)
        if waveunit is None:
            waveunit = default_waveunit
        unit = None
        if waveunit is not None:
            try:
                unit = u.Unit(waveunit)
            except ValueError:
                raise WaveunitNotConvertibleError(waveunit)
        for header_entry in entry.fits_header_entries:
            key, value = header_entry.key, header_entry.value
            if key == 'INSTRUME':
                entry.instrument = value
            elif key == 'WAVELNTH':
                if unit is None:
                    raise WaveunitNotFoundError(file)
                # use the value of `unit` to convert the wavelength to nm
                entry.wavemin = entry.wavemax = unit.to(
                    u.nm, value, equivalencies.spectral())
            # NOTE: the key DATE-END or DATE_END is not part of the official
            # FITS standard, but many FITS files use it in their header
            elif key in ('DATE-END', 'DATE_END'):
                try:
                    dt = parse_time(value).datetime
                except ValueError:
                    dt = Time.strptime(value, time_string_parse_format).datetime
                entry.observation_time_end = dt
            elif key in ('DATE-OBS', 'DATE_OBS'):
                try:
                    dt = parse_time(value).datetime
                except ValueError:
                    dt = Time.strptime(value, time_string_parse_format).datetime
                entry.observation_time_start = dt
        yield entry


def entries_from_dir(fitsdir, recursive=False, pattern='*',
                     default_waveunit=None, time_string_parse_format=None):
    """Search the given directory for FITS files and use the corresponding FITS
    headers to generate instances of :class:`DatabaseEntry`. FITS files are
    detected by reading the content of each file, the ``pattern`` argument may be
    used to avoid reading entire directories if one knows that all FITS files
    have the same filename extension.

    Parameters
    ----------
    fitsdir : str
        The directory where to look for FITS files.
    recursive : bool, optional
        If True, the given directory will be searched recursively. Otherwise,
        only the given directory and no subdirectories are searched. The
        default is `False`, i.e. the given directory is not searched
        recursively.
    pattern : str, optional
        The pattern can be used to filter the list of filenames before the
        files are attempted to be read. The default is to collect all files.
        This value is passed to the function :func:`fnmatch.filter`, see its
        documentation for more information on the supported syntax.
    default_waveunit : str, optional
        The wavelength unit that is used for a header if it cannot be found.
    time_string_parse_format : str, optional
        Fallback timestamp format which will be passed to
        `~astropy.time.Time.strptime` if `sunpy.time.parse_time` is unable to
        automatically read the ``date-obs`` metadata.

    Returns
    -------
    generator of (DatabaseEntry, str) pairs
        A generator where each item is a tuple consisting of a
        :class:`DatabaseEntry` instance and the absolute path to the filename
        which was used to make the database entry.

    Examples
    --------
    >>> import os
    >>> from sunpy.data.test import rootdir as fitsdir
    >>> from sunpy.database.tables import entries_from_dir

    >>> eitdir = os.path.join(fitsdir, 'EIT')
    >>> entries = list(entries_from_dir(eitdir, default_waveunit='angstrom'))
    >>> len(entries)
    13

    """
    for dirpath, dirnames, filenames in os.walk(fitsdir):
        filename_paths = (os.path.join(dirpath, name) for name in sorted(filenames))
        for path in fnmatch.filter(filename_paths, pattern):
            try:
                filetype = sunpy_filetools._detect_filetype(path)
            except (
                    sunpy_filetools.UnrecognizedFileTypeError,
                    sunpy_filetools.InvalidJPEG2000FileExtension):
                continue
            if filetype == 'fits':
                for entry in entries_from_file(
                        path, default_waveunit,
                        time_string_parse_format=time_string_parse_format
                ):
                    yield entry, path
        if not recursive:
            break


def _create_display_table(database_entries, columns=None, sort=False):
    """Generate a table to display the database entries.

    Parameters
    ----------
    database_entries : list
        The :class:`DatabaseEntry`s will be the rows in the resulting table.
    columns : list
        The column name strings that will be displayed in the resulting table. Possible
        values for the strings are all attributes of :class:`DatabaseEntry`.
    sort : bool (optional)
        If True, sorts the entries before displaying them.

    Returns
    -------
    str
        An astropy table that can be printed on the console or written to a
        file.

    """
    if columns is None:
        columns = ['id', 'observation_time_start', 'observation_time_end',
                   'instrument', 'source', 'provider', 'physobs', 'wavemin',
                   'wavemax', 'path', 'fileid', 'tags', 'starred',
                   'download_time', 'size']

    data = []
    for entry in database_entries:
        row = []
        for col in columns:
            if col == 'starred':
                row.append('Yes' if entry.starred else 'No')
            elif col == 'tags':
                row.append(', '.join(map(str, entry.tags)) or 'N/A')
            elif col == 'hdu_index':
                row.append(entry.hdu_index)
            # do not display microseconds in datetime columns
            elif col in (
                    'observation_time_start',
                    'observation_time_end',
                    'download_time'):
                time = getattr(entry, col, None)
                if time is None:
                    formatted_time = 'N/A'
                else:
                    formatted_time = time.strftime(TIME_FORMAT)
                row.append(formatted_time)
            else:
                row.append(str(getattr(entry, col) or 'N/A'))
        if not row:
            raise TypeError('at least one column must be given')
        data.append(row)
    if not data:
        raise TypeError('given iterable is empty')
    if sort:
        data.sort()
    return astropy.table.Table(rows=data, names=columns)


def display_entries(database_entries, columns=None, sort=False):
    """
    Print a table to display the database entries.

    Parameters
    ----------
    database_entries : iterable of `DatabaseEntry` instances
        The database entries will be the rows in the resulting table.
    columns : iterable of `str`
        The columns that will be displayed in the resulting table. Possible
        values for the strings are all attributes of :class:`DatabaseEntry`.
    sort : bool, optional
        If True, sorts the entries before displaying them.
    """
    return _create_display_table(database_entries, columns, sort).__str__()



import numpy as np
from matplotlib import patches
from matplotlib.path import Path

import astropy.units as u
from astropy.constants import R_sun
from astropy.coordinates import SkyCoord

from sunpy.coordinates.frames import HeliocentricInertial, Helioprojective
from sunpy.coordinates.sun import _angular_radius
from sunpy.coordinates.utils import get_limb_coordinates
from sunpy.visualization import wcsaxes_compat

__all__ = ['draw_limb']


@u.quantity_input
def draw_limb(axes, observer, *, rsun: u.m = R_sun, resolution=1000, **kwargs):
    """
    Draws the solar limb as seen by the specified observer.

    The limb is a circle for only the simplest plots.  If the specified
    observer of the limb is different from the observer of the coordinate frame
    of the plot axes, not only may the limb not be a true circle, a portion of
    the limb may be hidden from the observer.  In that case, the circle is
    divided into visible and hidden segments, represented by solid and dotted
    lines, respectively.

    Parameters
    ----------
    axes : `~matplotlib.axes` or ``None``
        Axes to plot limb on.
    observer : `astropy.coordinates.SkyCoord`
        Observer coordinate for which the limb is drawn.
    rsun : `~astropy.units.Quantity`
        Solar radius (in physical length units) at which to draw the limb.
        Defaults to the standard photospheric radius.
    resolution : `int`
        The number of points to use to represent the limb.

    Returns
    -------
    visible : `~matplotlib.patches.Polygon` or `~matplotlib.patches.Circle` or None
        The patch added to the axes for the visible part of the limb (i.e., the
        "near" side of the Sun).
    hidden : `~matplotlib.patches.Polygon` or None
        The patch added to the axes for the hidden part of the limb (i.e., the
        "far" side of the Sun).

    Notes
    -----
    Keyword arguments are passed onto the patches.

    If the limb is a true circle, ``visible`` will instead be
    `~matplotlib.patches.Circle` and ``hidden`` will be ``None``.

    If there are no hidden points (e.g., on a synoptic map any limb is fully
    visible) ``hidden`` will be ``None``.

    If there are no visible points (e.g., for an observer on the opposite side
    of the Sun to the map observer) ``visible`` will be ``None``.

    To avoid triggering Matplotlib auto-scaling, these patches are added as
    artists instead of patches.  One consequence is that the plot legend is not
    populated automatically when the limb is specified with a text label.  See
    :ref:`sphx_glr_gallery_text_labels_and_annotations_custom_legends.py` in
    the Matplotlib documentation for examples of creating a custom legend.
    """
    if not wcsaxes_compat.is_wcsaxes(axes):
        raise ValueError('axes must be a WCSAxes')

    c_kw = {'fill': False,
            'color': 'white',
            'zorder': 100}
    c_kw.update(kwargs)

    transform = axes.get_transform('world')
    # transform is always passed on as a keyword argument
    c_kw.setdefault('transform', transform)

    # If the observer matches the axes's observer frame and is Helioprojective, use a Circle
    axes_frame = axes._transform_pixel2world.frame_out
    if isinstance(axes_frame, Helioprojective):
        axes_observer = SkyCoord(axes_frame.observer)
        if axes_observer.separation_3d(observer) < 1 * u.m:
            distance = observer.transform_to(HeliocentricInertial).spherical.distance
            # Obtain the solar radius and the world->pixel transform
            angular_radius = _angular_radius(rsun, distance)
            circ = patches.Circle([0, 0], radius=angular_radius.to_value(u.deg), **c_kw)
            axes.add_artist(circ)
            return circ, None

    # Otherwise, we use Polygon to be able to distort the limb
    # Create the limb coordinate array using Heliocentric Radial
    limb = get_limb_coordinates(observer, rsun, resolution)

    # Transform the limb to the axes frame and get the 2D vertices
    limb_in_axes = limb.transform_to(axes_frame)
    Tx = limb_in_axes.spherical.lon.to_value(u.deg)
    Ty = limb_in_axes.spherical.lat.to_value(u.deg)
    vertices = np.array([Tx, Ty]).T

    # Determine which points are visible
    if hasattr(axes_frame, 'observer'):
        # The reference distance is the distance to the limb for the axes
        # observer
        rsun = getattr(axes_frame, 'rsun', rsun)
        reference_distance = np.sqrt(axes_frame.observer.radius**2 - rsun**2)
        is_visible = limb_in_axes.spherical.distance <= reference_distance
    else:
        # If the axes has no observer, the entire limb is considered visible
        is_visible = np.ones_like(
            limb_in_axes.spherical.distance, bool, subok=False)

    # Identify discontinuities in the limb. Uses the same approach as
    # astropy.visualization.wcsaxes.grid_paths.get_lon_lat_path()
    step = np.sqrt((vertices[1:, 0] - vertices[:-1, 0]) ** 2 +
                   (vertices[1:, 1] - vertices[:-1, 1]) ** 2)
    continuous = np.concatenate([[True, True], step[1:] < 100 * step[:-1]])

    visible, hidden = None, None
    if np.sum(is_visible) > 0:
        # Create the Polygon for the near side of the Sun (using a solid line)
        if 'linestyle' not in kwargs:
            c_kw['linestyle'] = '-'
        visible = patches.Polygon(vertices, **c_kw)
        _modify_polygon_visibility(visible, is_visible & continuous)
        # Add patches as artists rather than patches to avoid triggering auto-scaling
        axes.add_artist(visible)

    if np.sum(~is_visible) > 0:
        # Create the Polygon for the far side of the Sun (using a dotted line)
        if 'linestyle' not in kwargs:
            c_kw['linestyle'] = ':'
        hidden = patches.Polygon(vertices, **c_kw)
        _modify_polygon_visibility(hidden, ~is_visible & continuous)
        axes.add_artist(hidden)

    return visible, hidden


def _modify_polygon_visibility(polygon, keep):
    polygon_codes = polygon.get_path().codes
    polygon_codes[:-1][~keep] = Path.MOVETO
    polygon_codes[-1] = Path.MOVETO if not keep[0] else Path.LINETO



import importlib

from sunpy.util.sysinfo import missing_dependencies_by_extra


def print_missing_dependencies_report(missing, package="sunpy"):
    printed = False
    required_missing = missing.pop("required")
    if required_missing:
        printed = True
        print(f"The following packages are not installed but are required by {package}:")
        for dep in required_missing:
            print(f"* {dep}")
    for extra_name, dependencies in missing.items():
        if not dependencies:
            continue
        printed = True
        print(
            f"The following packages are not installed for the {package}[{extra_name}] requirement:")
        for dep in dependencies:
            print(f"  * {dep}")
    return printed


def _self_test_args(*, package=None, online=False, online_only=False, figure_only=False):
    args = ["-W", "ignore"]
    if online:
        args.append("--remote-data=any")
    if online_only:
        args.append("--remote-data=any -m remote_data")
    if package:
        try:
            importlib.import_module(f"sunpy.{package}")
        except ModuleNotFoundError:
            raise ModuleNotFoundError(f"sunpy.{package} was not found.")
        args.extend(["--pyargs", f"sunpy.{package}"])
    else:
        args.extend(["--pyargs", "sunpy"])
    if figure_only:
        args.extend(["-m", "mpl_image_compare"])
    return args


def self_test(*, package=None, online=False, online_only=False, figure_only=False):
    print("\n\n")
    print("Starting sunpy self test...")
    print("Checking for packages needed to run sunpy:")
    missing = missing_dependencies_by_extra(exclude_extras=["asdf", "dask", "dev", "all", "docs"])
    test_missing = missing.pop("tests")
    printed = print_missing_dependencies_report(missing)
    if not printed:
        print("All required and optional sunpy dependencies are installed.")
    if test_missing:
        print("You do not have all the required dependencies installed to run the sunpy test suite.")
        print(list(test_missing.keys()))
        print("If are using conda, you will want to run `conda install <package name>`")
        print('Otherwise you will want run `pip install "sunpy[all,tests]"`')
        return
    import pytest
    print("Starting the sunpy test suite:")
    print()
    args = _self_test_args(package=package, online=online,
                           online_only=online_only, figure_only=figure_only)
    return pytest.main(args)



"""Test cases for SDO Map subclasses.
This particular test file pertains to AIAMap.
@Author: Pritish C. (VaticanCameos)
"""
import os
import glob

import pytest

import astropy.units as u
from astropy.visualization.mpl_normalize import ImageNormalize

import sunpy.data.test
from sunpy.map import Map
from sunpy.map.sources.sdo import AIAMap
from sunpy.tests.helpers import SKIP_GLYMUR

path = sunpy.data.test.rootdir
jp2path = glob.glob(os.path.join(path, "2013_06_24__17_31_30_84__SDO_AIA_AIA_193.jp2"))
aiaimg = glob.glob(os.path.join(path, "aia_171_level1.fits"))


if SKIP_GLYMUR:
    params = [aiaimg]
else:
    params = [aiaimg, jp2path]


# The fixture is parameterized with aiaimg and jp2path.
@pytest.fixture(scope="module", params=params)
def createAIAMap(request):
    """Creates an AIAMap as given in documentation examples, through AIA_171_IMAGE
    or through the use of the JP2 file."""
    aiaobj = Map(request.param)
    return aiaobj

# AIA Tests


def test_AIAMap(createAIAMap):
    """Tests the creation of AIAMap from AIA_171_IMAGE or through
    use of the JP2 file."""
    assert isinstance(createAIAMap, AIAMap)


def test_is_datasource_for(createAIAMap):
    """Tests the is_datasource_for method of AIAMap."""
    assert createAIAMap.is_datasource_for(createAIAMap.data, createAIAMap.meta)


def test_observatory(createAIAMap):
    """Tests the observatory property of the AIAMap object."""
    assert createAIAMap.observatory == "SDO"


def test_measurement(createAIAMap):
    """Tests the measurement property of the AIAMap object."""
    assert createAIAMap.measurement.value in [171, 193]
    # aiaimg has 171, jp2path has 193.


def test_norm_clip(createAIAMap):
    # Tests that the default normalizer has clipping disabled
    assert not createAIAMap.plot_settings['norm'].clip


def test_new_instance_preserves_plot_settings(createAIAMap):
    # Tests that the _new_instance method preserves the plot_settings
    # of the old instance. This is done on the AIA source as the AIAMap
    # constructor explicitly sets the cmap and norm and we want to test
    # that _new_instance persists the old custom plot_settings
    createAIAMap.plot_settings['norm'] = ImageNormalize(vmin=0.1, vmax=42)
    createAIAMap.plot_settings['cmap'] = 'inferno'
    new_aia_map = createAIAMap._new_instance(createAIAMap.data,
                                             createAIAMap.meta,
                                             plot_settings=createAIAMap.plot_settings)
    assert new_aia_map.plot_settings['norm'].vmin == createAIAMap.plot_settings['norm'].vmin
    assert new_aia_map.plot_settings['norm'].vmax == createAIAMap.plot_settings['norm'].vmax
    assert new_aia_map.plot_settings['cmap'] == createAIAMap.plot_settings['cmap']
    # If no plot settings are explicitly passed, the plot_settings should fall back to those
    # in the constructor
    new_aia_map = createAIAMap._new_instance(createAIAMap.data, createAIAMap.meta)
    assert new_aia_map.plot_settings['norm'].vmin is None
    assert new_aia_map.plot_settings['norm'].vmax is None
    assert new_aia_map.plot_settings['cmap'] == new_aia_map._get_cmap_name()


def test_wcs(createAIAMap):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    createAIAMap.pixel_to_world(0*u.pix, 0*u.pix)



"""
This module implements a SRS File Reader.
"""
import datetime
from collections import OrderedDict

import numpy as np

import astropy.io.ascii
import astropy.units as u
from astropy.table import Column, MaskedColumn, QTable, vstack

__all__ = ['read_srs']


def read_srs(filepath):
    """
    Parse a SRS table from NOAA SWPC.

    Parameters
    ----------
    filepath : `str`
        The full path to a SRS table.

    Returns
    -------
    table : `astropy.table.QTable`
        Table containing a stacked table from all the tables in the SRS file.
        The header information is stored in the ``.meta`` attribute.
    """
    with open(filepath) as srs:
        file_lines = srs.readlines()

    header, section_lines = split_lines(file_lines)

    return make_table(header, section_lines)


def make_table(header, section_lines):
    """
    From the seperated section lines and the header, clean up the data and
    convert to a `~astropy.table.QTable`.
    """
    meta_data = get_meta_data(header)

    tables = []
    for i, lines in enumerate(section_lines):
        if lines:
            key = list(meta_data['id'].keys())[i]
            t1 = astropy.io.ascii.read(lines)

            if len(t1) == 0:
                col_data_types = {
                    # ID : <class 'str'>
                    'Nmbr': np.dtype('i4'),
                    'Location': np.dtype('U6'),
                    'Lo': np.dtype('i8'),
                    'Area': np.dtype('i8'),
                    'Z': np.dtype('U3'),
                    'LL': np.dtype('i8'),
                    'NN': np.dtype('i8'),
                    'MagType': np.dtype('S4'),
                    'Lat': np.dtype('i8')
                }
                for c in t1.itercols():
                    # Put data types of columns in empty table to correct types,
                    # or else vstack will fail.
                    c.dtype = col_data_types[c._name]
                t1.add_column(
                    Column(data=None, name="ID", dtype=('S2')), index=0)
            else:
                t1.add_column(Column(data=[key] * len(t1), name="ID"), index=0)

            tables.append(t1)

    out_table = vstack(tables)

    # Parse the Location column in Table 1
    if 'Location' in out_table.columns:
        col_lat, col_lon = parse_location(out_table['Location'])
        del out_table['Location']
        out_table.add_column(col_lat)
        out_table.add_column(col_lon)

    # Parse the Lat column in Table 3
    if 'Lat' in out_table.columns:
        parse_lat_col(out_table['Lat'], out_table['Latitude'])
        del out_table['Lat']

    # Give columns more sensible names
    out_table.rename_column("Nmbr", "Number")
    out_table.rename_column("NN", "Number of Sunspots")
    out_table.rename_column("Lo", "Carrington Longitude")
    out_table.rename_column("MagType", "Mag Type")
    out_table.rename_column("LL", "Longitudinal Extent")

    # Define a Solar Hemispere Unit
    a = {}
    u.def_unit(
        "SH",
        represents=(2 * np.pi * u.solRad**2),
        prefixes=True,
        namespace=a,
        doc="A solar hemisphere is the area of the visible solar disk.")

    # Set units on the table
    out_table['Carrington Longitude'].unit = u.deg
    out_table['Area'].unit = a['uSH']
    out_table['Longitudinal Extent'].unit = u.deg

    out_table.meta = meta_data

    # Number should be formatted in 10000 after 2002-06-15.
    if out_table.meta['issued'] > datetime.datetime(2002, 6, 15):
        out_table['Number'] += 10000

    return QTable(out_table)


def split_lines(file_lines):
    """
    Given all the lines in the file split based on the three sections and
    return the lines for the header and a list of lines for each section that
    is not 'None'.
    """
    section_lines = []
    for i, line in enumerate(file_lines):
        if line.startswith(("I.", "IA.", "II.")):
            section_lines.append(i)

    header = file_lines[:section_lines[0]]
    header += [file_lines[s] for s in section_lines]

    # Append comments to the comment lines
    for l in section_lines:
        file_lines[l] = '# ' + file_lines[l]
    t1_lines = file_lines[section_lines[0]:section_lines[1]]
    # Remove the space so table reads it correctly
    t1_lines[1] = t1_lines[1].replace('Mag Type', 'MagType')
    t2_lines = file_lines[section_lines[1]:section_lines[2]]
    t3_lines = file_lines[section_lines[2]:]

    lines = [t1_lines, t2_lines, t3_lines]
    for i, ll in enumerate(lines):
        if ll[2].strip() == 'None':
            del ll[2]

    return header, lines


def get_meta_data(header):
    """
    Convert a list of header lines into a meta data dict.
    """
    meta_lines = []
    for l in header:
        if l.startswith(':'):
            meta_lines.append(l)

    meta_data = {}
    for m in meta_lines:
        if "Corrected Copy" in m:
            meta_data['corrected'] = True
            continue
        k, v = m.strip().split(':')[1:]
        meta_data[k.lower()] = v.strip()
    meta_data['issued'] = datetime.datetime.strptime(meta_data['issued'],
                                                     "%Y %b %d %H%M UTC")

    # Get ID descriptions
    meta_data['id'] = OrderedDict()
    for h in header:
        if h.startswith(("I.", "IA.", "II.")):
            i = h.find('.')
            k = h[:i]
            v = h[i + 2:]
            meta_data['id'][k] = v.strip()

    meta_data['header'] = [h.strip() for h in header]
    return meta_data


def parse_longitude(value):
    """
    Parse longitude in the form "W10" or "E10".
    """
    lonsign = {'W': 1, 'E': -1}
    if "W" in value or "E" in value:
        return lonsign[value[3]] * float(value[4:])


def parse_latitude(value):
    """
    Parse latitude in the form "S10" or "N10".
    """
    latsign = {'N': 1, 'S': -1}
    if "N" in value or "S" in value:
        return latsign[value[0]] * float(value[1:3])


def parse_location(column):
    """
    Given a column of location data in the form "S10E10" convert to two columns
    of angles.
    """
    latitude = MaskedColumn(name="Latitude", unit=u.deg)
    longitude = MaskedColumn(name="Longitude", unit=u.deg)

    for i, loc in enumerate(column):
        if loc:
            lati = parse_latitude(loc)
            longi = parse_longitude(loc)
            latitude = latitude.insert(i, lati)
            longitude = longitude.insert(i, longi)
        else:
            latitude = latitude.insert(i, None, mask=True)
            longitude = longitude.insert(i, None, mask=True)
    return latitude, longitude


def parse_lat_col(column, latitude_column):
    """
    Given an input column of "latitudes" in the form "S10" parse them and add
    them to an existing column of "latitudes".
    """
    for i, loc in enumerate(column):
        if loc:
            latitude_column.mask[i] = False
            latitude_column[i] = parse_latitude(loc)
    return latitude_column



import warnings

import numpy as np
import pytest
from numpy.testing import assert_allclose

import astropy.units as u
from astropy.coordinates import Angle, EarthLocation
from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time
from astropy.utils.exceptions import ErfaWarning

from sunpy.coordinates import sun
from sunpy.sun.constants import radius
from .helpers import assert_longitude_allclose

# Ignore warnings that astropy throws when trying and failing to update ephemeris
pytestmark = pytest.mark.filterwarnings("ignore:failed to download")


@pytest.fixture
def t1():
    return Time('1992-10-13', scale='tdb')


@pytest.fixture
def t2():
    return Time('2013-06-17', scale='tt')


def test_true_longitude(t1, t2):
    # Validate against a published value from the Astronomical Almanac (1992, C16)
    assert_quantity_allclose(sun.true_longitude(t1), Angle('199d54m26.17s'), atol=0.005*u.arcsec)

    # Validate against a published value from the Astronomical Almanac (2013, C12)
    assert_quantity_allclose(sun.true_longitude(t2), Angle('85d58m57.46s'), atol=0.005*u.arcsec)


def test_apparent_longitude(t1, t2):
    aberration_1au = Angle('20.496s')
    # Validate against a published value from the Astronomical Almanac (1992, C2+C16+B30)
    assert_quantity_allclose(sun.apparent_longitude(t1),
                             Angle('199d54m26.17s') + Angle('15.908s') - aberration_1au / 0.9976085,
                             atol=0.005*u.arcsec)

    # Validate against a published value from the Astronomical Almanac (2013, C2+C12+B61)
    assert_quantity_allclose(sun.apparent_longitude(t2),
                             Angle('85d58m57.46') + Angle('12.0489s') - aberration_1au / 1.0159149,
                             atol=0.005*u.arcsec)


def test_true_latitude(t1, t2):
    # Validate against a published value from the Astronomical Almanac (1992, C16)
    # Disagreement with published precision may be due to changes in definitions after 1992
    assert_quantity_allclose(sun.true_latitude(t1), Angle('0.72s'), atol=0.05*u.arcsec)

    # Validate against a published value from the Astronomical Almanac (2013, C12)
    assert_quantity_allclose(sun.true_latitude(t2), Angle('-0.42s'), atol=0.005*u.arcsec)


def test_apparent_latitude(t1, t2):
    # Validate against a published value from the Astronomical Almanac (1992, C2+C16)
    # Disagreement with published precision may be due to changes in definitions after 1992
    assert_quantity_allclose(sun.apparent_latitude(t1), Angle('0.72s'), atol=0.05*u.arcsec)

    # Validate against a published value from the Astronomical Almanac (2013, C2+C12)
    assert_quantity_allclose(sun.apparent_latitude(t2), Angle('-0.42s'), atol=0.005*u.arcsec)


def test_angular_radius(t2):
    # Validate against a published value from the Astronomical Almanac (2013, C13)
    # The Astromomical Almanac uses a slightly different radius for the Sun (6.96e5 km)
    # The Astronomical Almanac also uses a small-angle approximation
    # See https://archive.org/details/131123ExplanatorySupplementAstronomicalAlmanac/page/n212/mode/1up
    assert_quantity_allclose(sun.angular_radius(t2),
                             Angle('0d15m44.61s') / (6.96e5*u.km) * radius,  # scale to IAU radius
                             atol=0.005*u.arcsec)

    # Regression-only test
    assert_quantity_allclose(sun.angular_radius("2012/11/11"), 968.875*u.arcsec, atol=1e-3*u.arcsec)
    with pytest.warns(ErfaWarning):
        assert_quantity_allclose(sun.angular_radius("2043/03/01"),
                                 968.330*u.arcsec, atol=1e-3*u.arcsec)
    assert_quantity_allclose(sun.angular_radius("2001/07/21"), 944.042*u.arcsec, atol=1e-3*u.arcsec)


def test_mean_obliquity_of_ecliptic(t1, t2):
    # Validate against a published value from the Astronomical Almanac (1992, B30)
    # Note that the publication date pre-dates the IAU 2006 definition of obliquity
    assert_quantity_allclose(sun.mean_obliquity_of_ecliptic(t1),
                             Angle('23d26m24.519s') + 0.308*u.arcsec, atol=0.05*u.arcsec)

    # Validate against a published value from the Astronomical Almanac (2013, B61)
    assert_quantity_allclose(sun.mean_obliquity_of_ecliptic(t2),
                             Angle('23d26m08.0875s') + 7.0153*u.arcsec, atol=0.00005*u.arcsec)


def test_true_rightascension():
    # Regression-only test
    assert_quantity_allclose(sun.true_rightascension("2012/11/11"), 226.548*u.deg, atol=1e-3*u.deg)
    with pytest.warns(ErfaWarning):
        assert_quantity_allclose(sun.true_rightascension(
            "2142/02/03"), 316.466*u.deg, atol=1e-3*u.deg)
    assert_quantity_allclose(sun.true_rightascension("2013/12/11"), 258.150*u.deg, atol=1e-3*u.deg)


def test_true_rightascension_J2000(t1, t2):
    # Validate against JPL HORIZONS output
    # https://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&TABLE_TYPE=OBSERVER&COMMAND=10&CENTER=500&QUANTITIES=1&ANG_FORMAT=HMS&EXTRA_PREC=YES&TIME_TYPE=TT&TLIST=2448908.5%0A2456460.5
    assert_quantity_allclose(sun.true_rightascension(t1, equinox_of_date=False),
                             Angle('13h13m53.65s'), atol=Angle('0h0m0.005s'))
    assert_quantity_allclose(sun.true_rightascension(t2, equinox_of_date=False),
                             Angle('05h41m40.32s'), atol=Angle('0h0m0.005s'))


def test_true_declination():
    # Regression-only test
    assert_quantity_allclose(sun.true_declination("2012/11/11"), -17.470*u.deg, atol=1e-3*u.deg)
    with pytest.warns(ErfaWarning):
        assert_quantity_allclose(sun.true_declination("2245/12/01"), -21.717*u.deg, atol=1e-3*u.deg)
    assert_quantity_allclose(sun.true_declination("2014/05/27"), 21.245*u.deg, atol=1e-3*u.deg)


def test_true_declination_J2000(t1, t2):
    # Validate against JPL HORIZONS output
    # https://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&TABLE_TYPE=OBSERVER&COMMAND=10&CENTER=500&QUANTITIES=1&ANG_FORMAT=HMS&EXTRA_PREC=YES&TIME_TYPE=TT&TLIST=2448908.5%0A2456460.5
    assert_quantity_allclose(sun.true_declination(t1, equinox_of_date=False),
                             Angle('-7d49m20.84s'), atol=0.005*u.arcsec)
    assert_quantity_allclose(sun.true_declination(t2, equinox_of_date=False),
                             Angle('23d22m13.97s'), atol=0.005*u.arcsec)


def test_true_obliquity_of_ecliptic(t1, t2):
    # Validate against a published value from the Astronomical Almanac (1992, B30)
    # Note that the publication date pre-dates the IAU 2006 definition of obliquity
    assert_quantity_allclose(sun.true_obliquity_of_ecliptic(t1), Angle('23d26m24.519s'),
                             atol=0.05*u.arcsec)

    # Validate against a published value from the Astronomical Almanac (2013, B61)
    assert_quantity_allclose(sun.true_obliquity_of_ecliptic(t2), Angle('23d26m08.0875'),
                             atol=0.00005*u.arcsec)


def test_apparent_rightascension(t1, t2):
    # Validate against a published value from the Astronomical Almanac (1992, C16)
    assert_quantity_allclose(sun.apparent_rightascension(t1), Angle('13h13m30.75s'),
                             atol=Angle('0h0m0.005s'))

    # Validate against a published value from the Astronomical Almanac (2013, C12)
    assert_quantity_allclose(sun.apparent_rightascension(t2), Angle('5h42m28.88s'),
                             atol=Angle('0h0m0.005s'))


def test_apparent_rightascension_J2000(t1):
    # Regression-only test
    assert_quantity_allclose(sun.apparent_rightascension(t1, equinox_of_date=False),
                             Angle('13h13m52.37s'), atol=Angle('0h0m0.005s'))


def test_apparent_declination(t1, t2):
    # Validate against a published value from the Astronomical Almanac (1992, C16)
    assert_quantity_allclose(sun.apparent_declination(t1), Angle('-7d47m01.7s'), atol=0.05*u.arcsec)

    # Validate against a published value from the Astronomical Almanac (2013, C12)
    assert_quantity_allclose(sun.apparent_declination(t2), Angle('23d22m27.8s'), atol=0.05*u.arcsec)


def test_apparent_declination_J2000(t1):
    # Regression-only test
    assert_quantity_allclose(sun.apparent_declination(t1, equinox_of_date=False),
                             Angle('-7d49m13.1s'), atol=0.05*u.arcsec)


def test_sky_position(t1, t2):
    pos1 = sun.sky_position(t1)
    ra1 = sun.apparent_rightascension(t1)
    dec1 = sun.apparent_declination(t1)
    assert_quantity_allclose(pos1, (ra1, dec1))

    pos2 = sun.sky_position(t2, equinox_of_date=False)
    ra2 = sun.apparent_rightascension(t2, equinox_of_date=False)
    dec2 = sun.apparent_declination(t2, equinox_of_date=False)
    assert_quantity_allclose(pos2, (ra2, dec2))


@pytest.mark.filterwarnings('ignore:Tried to get polar motions for times after IERS data is valid')
@pytest.mark.filterwarnings(r'ignore:\(some\) times are outside of range covered by IERS table')
def test_print_params():
    # Test only for any issues with printing; accuracy is covered by other tests
    sun.print_params()


def test_B0():
    # Validate against a published value from Astronomical Algorithms (Meeus 1998, p.191)
    assert_quantity_allclose(sun.B0('1992-Oct-13'), 5.99*u.deg, atol=5e-3*u.deg)


def test_B0_astronomical_almanac():
    # Validate against published values from the Astronomical Almanac (2013)
    sun_B0 = sun.B0(Time(['2013-04-01', '2013-12-01'], scale='tt'))
    assert_quantity_allclose(sun_B0[0], -6.54*u.deg, atol=5e-3*u.deg)
    assert_quantity_allclose(sun_B0[1], 0.88*u.deg, atol=5e-3*u.deg)


def test_B0_jpl_horizons():
    # Validate against values from JPL Horizons, which does not apply the aberration correction
    # for observer motion.
    # https://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&TABLE_TYPE=OBSERVER&OBJ_DATA=NO&QUANTITIES=%2714%27&COMMAND=%22Sun%22&CENTER=%27Geocentric%27&START_TIME=%222013-01-01+TT%22&STOP_TIME=%222013-12-31%22&STEP_SIZE=%221d%22
    jpl_values = {'2013-01-01': -3.034687,
                  '2013-02-01': -6.031910,
                  '2013-03-01': -7.219984,
                  '2013-04-01': -6.543092,
                  '2013-05-01': -4.168016,
                  '2013-06-01': -0.663804,
                  '2013-07-01': +2.873030,
                  '2013-08-01': +5.784693,
                  '2013-09-01': +7.194559,
                  '2013-10-01': +6.719875,
                  '2013-11-01': +4.378979,
                  '2013-12-01': +0.881068}
    sun_B0 = sun.B0(Time(list(jpl_values.keys()), scale='tt'))
    assert_quantity_allclose(sun_B0, list(jpl_values.values()) * u.deg, atol=0.01*u.arcsec, rtol=0)


def test_B0_sunspice():
    # Validate against values from SunSPICE (including calling CSPICE functions)
    # Without the aberration correction for observer motion (specify 'LT')
    #
    # IDL> load_sunspice_gen
    # IDL> cspice_str2et, '2013-01-01', et
    # IDL> cspice_subpnt, 'Intercept/Ellipsoid', 'Sun', et, 'IAU_Sun', 'LT', 'Earth', spoint, trgepc, srfvec
    # IDL> print, spclat * cspice_dpr()
    #       -3.0347784
    jpl_values = {'2013-01-01': -3.0347784,
                  '2013-02-01': -6.0319658,
                  '2013-03-01': -7.2199937,
                  '2013-04-01': -6.5430498,
                  '2013-05-01': -4.1679382,
                  '2013-06-01': -0.66371004,
                  '2013-07-01': +2.8731155,
                  '2013-08-01': +5.7847503,
                  '2013-09-01': +7.1945709,
                  '2013-10-01': +6.7198381,
                  '2013-11-01': +4.3789008,
                  '2013-12-01': +0.88096965}
    sun_B0 = sun.B0(Time(list(jpl_values.keys()), scale='utc'))
    assert_quantity_allclose(sun_B0, list(jpl_values.values()) * u.deg, atol=0.005*u.arcsec, rtol=0)


def test_L0_defaults():
    # Confirm that the output is the same when the default settings are explicitly set
    defaults = sun.L0('1992-Oct-13')
    explicit = sun.L0('1992-Oct-13',
                      light_travel_time_correction=True,
                      aberration_correction=False,
                      nearest_point=True)
    assert defaults == explicit


def test_L0_astronomical_almanac():
    # Validate against published values from the Astronomical Almanac (2013)
    aa_values = {'2013-01-01': 326.98,
                 '2013-02-01': 278.78,
                 '2013-03-01': 270.07,
                 '2013-04-01': 221.44,
                 '2013-05-01': 185.30,
                 '2013-06-01': 135.30,
                 '2013-07-01':  98.22,  # NOQA
                 '2013-08-01':  48.03,  # NOQA
                 '2013-09-01': 358.28,
                 '2013-10-01': 322.22,
                 '2013-11-01': 273.31,
                 '2013-12-01': 237.83}
    sun_L0 = sun.L0(Time(list(aa_values.keys()), scale='tt'),
                    light_travel_time_correction=True,
                    aberration_correction=True,
                    nearest_point=False)
    assert_longitude_allclose(sun_L0, list(aa_values.values()) * u.deg, atol=5e-3*u.deg)


def test_L0_jpl_horizons():
    # Validate against values from JPL Horizons, which does not apply the aberration correction
    # for observer motion.
    # https://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&TABLE_TYPE=OBSERVER&OBJ_DATA=NO&QUANTITIES=%2714%27&COMMAND=%22Sun%22&CENTER=%27Geocentric%27&START_TIME=%222013-01-01+TT%22&STOP_TIME=%222013-12-31%22&STEP_SIZE=%221d%22
    jpl_values = {'2013-01-01': 326.989970,
                  '2013-02-01': 278.789894,
                  '2013-03-01': 270.072186,
                  '2013-04-01': 221.440599,
                  '2013-05-01': 185.306476,
                  '2013-06-01': 135.303097,
                  '2013-07-01':  98.221806,  # NOQA
                  '2013-08-01':  48.035951,  # NOQA
                  '2013-09-01': 358.289921,
                  '2013-10-01': 322.226009,
                  '2013-11-01': 273.315206,
                  '2013-12-01': 237.836959}
    sun_L0 = sun.L0(Time(list(jpl_values.keys()), scale='tt'),
                    light_travel_time_correction=True,
                    aberration_correction=False,
                    nearest_point=True)
    assert_longitude_allclose(sun_L0, list(jpl_values.values()) * u.deg, atol=0.01*u.arcsec)


def test_L0_sunspice():
    # Validate against values from SunSPICE (including calling CSPICE functions)

    # With the aberration correction for observer motion (specify 'LT+S')
    #
    # IDL> load_sunspice_gen
    # IDL> cspice_str2et, '2013-01-01', et
    # IDL> cspice_subpnt, 'Intercept/Ellipsoid', 'Sun', et, 'IAU_Sun', 'LT+S', 'Earth', spoint1, trgepc1, srfvec1
    # IDL> cspice_reclat, spoint1, spcrad1, spclon1, spclat1
    # IDL> print, spclon1 * cspice_dpr()
    #       -33.025998
    values1 = {'2013-01-01': -33.025998,
               '2013-02-01': -81.226108,
               '2013-03-01': -89.943817,
               '2013-04-01': -138.57536,
               '2013-05-01': -174.70941,
               '2013-06-01': +135.28726,
               '2013-07-01': +98.205970,
               '2013-08-01': +48.020071,
               '2013-09-01': -1.7260099,
               '2013-10-01': -37.789945,
               '2013-11-01': -86.700744,
               '2013-12-01': -122.17899}
    sun_L0 = sun.L0(Time(list(values1.keys()), scale='utc'),
                    light_travel_time_correction=True,
                    aberration_correction=True,
                    nearest_point=True)
    assert_longitude_allclose(sun_L0, list(values1.values()) * u.deg, atol=0.3*u.arcsec)

    # Without the aberration correction for observer motion (specify 'LT')
    #
    # IDL> cspice_subpnt, 'Intercept/Ellipsoid', 'Sun', et, 'IAU_Sun', 'LT', 'Earth', spoint2, trgepc2, srfvec2
    # IDL> cspice_reclat, spoint2, spcrad2, spclon2, spclat2
    # IDL> print, spclon2 * cspice_dpr()
    #       -33.020271
    values2 = {'2013-01-01': -33.020271,
               '2013-02-01': -81.220344,
               '2013-03-01': -89.938057,
               '2013-04-01': -138.56966,
               '2013-05-01': -174.70380,
               # '2013-06-01':  135.29281,  # skipping due to low precision in comparison
               '2013-07-01': +98.211514,
               '2013-08-01': +48.025667,
               '2013-09-01': -1.7203500,
               '2013-10-01': -37.784252,
               '2013-11-01': -86.695047,
               '2013-12-01': -122.17329}
    sun_L0 = sun.L0(Time(list(values2.keys()), scale='utc'),
                    light_travel_time_correction=True,
                    aberration_correction=False,
                    nearest_point=True)
    assert_longitude_allclose(sun_L0, list(values2.values()) * u.deg, atol=0.01*u.arcsec)

    # Without any corrections (do a straight conversion from 'HEQ' to 'Carrington')
    #
    # IDL> coord = [1.d, 0.d, 10.d]
    # IDL> convert_sunspice_lonlat, '2013-01-01', coord, 'HEQ', 'Carrington', /au, /degrees
    # IDL> print, coord
    #        1.0000000       326.89956       10.000000
    values3 = {'2013-01-01': 326.89956,
               '2013-02-01': 278.69932,
               '2013-03-01': 269.98115,
               '2013-04-01': 221.34886,
               '2013-05-01': 185.21404,
               '2013-06-01': 135.21012,
               '2013-07-01': 98.128607,
               '2013-08-01': 47.942897,
               '2013-09-01': 358.19735,
               '2013-10-01': 322.13410,
               '2013-11-01': 273.22402,
               '2013-12-01': 237.74631}
    sun_L0 = sun.L0(Time(list(values3.keys()), scale='utc'),
                    light_travel_time_correction=False,
                    aberration_correction=False,
                    nearest_point=True)
    assert_longitude_allclose(sun_L0, list(values3.values()) * u.deg, atol=0.02*u.arcsec)


def test_P():
    # Validate against a published value from Astronomical Algorithms (Meeus 1998, p.191)
    assert_quantity_allclose(sun.P('1992-Oct-13'), 26.27*u.deg, atol=5e-3*u.deg)


def test_P_array_time():
    # Validate against published values from the Astronomical Almanac (2013)
    sun_P = sun.P(Time(['2013-01-01',
                        '2013-02-01',
                        '2013-03-01',
                        '2013-04-01',
                        '2013-05-01',
                        '2013-06-01',
                        '2013-07-01',
                        '2013-08-01',
                        '2013-09-01',
                        '2013-10-01',
                        '2013-11-01',
                        '2013-12-01'], scale='tt'))
    assert_quantity_allclose(sun_P, [1.98,
                                     -12.23,
                                     -21.55,
                                     -26.15,
                                     -24.11,
                                     -15.39,
                                     -2.64,
                                     10.85,
                                     21.08,
                                     25.97,
                                     24.47,
                                     16.05]*u.deg, atol=5e-3*u.deg)


def test_earth_distance():
    # Validate against a published value from Astronomical Algorithms (Meeus 1998, p.191)
    assert_quantity_allclose(sun.earth_distance('1992-Oct-13'), 0.997608*u.AU, atol=5e-7*u.AU)


def test_earth_distance_array_time():
    # Validate against published values from the Astronomical Almanac (2013)
    sunearth_distance = sun.earth_distance(Time(['2013-04-01', '2013-12-01'], scale='tt'))
    assert_quantity_allclose(sunearth_distance[0], 0.9992311*u.AU, rtol=0, atol=5e-8*u.AU)
    assert_quantity_allclose(sunearth_distance[1], 0.9861362*u.AU, rtol=0, atol=5e-8*u.AU)


def test_orientation():
    # Not currently aware of a published value to check against, so just self-check for now

    # Check the Northern Hemisphere
    angle = sun.orientation(EarthLocation(lat=40*u.deg, lon=-75*u.deg), '2017-07-18 12:00')
    assert_quantity_allclose(angle, -59.4*u.deg, atol=0.1*u.deg)

    # Check the Southern Hemisphere
    angle = sun.orientation(EarthLocation(lat=-40*u.deg, lon=-75*u.deg), '2017-02-18 13:00')
    assert_quantity_allclose(angle, -110.8*u.deg, atol=0.1*u.deg)


# Validate against published values from the Astronomical Almanac (2013, C4)
@pytest.mark.parametrize("date, day_fraction, rotation_number",
                         [('2012-12-29', 0.49, 2132),
                          ('2013-01-25', 0.83, 2133),
                          ('2013-02-22', 0.17, 2134),
                          ('2013-03-21', 0.49, 2135),
                          ('2013-04-17', 0.78, 2136),
                          ('2013-05-15', 0.02, 2137),
                          ('2013-06-11', 0.22, 2138),
                          ('2013-07-08', 0.42, 2139),
                          ('2013-08-04', 0.63, 2140),
                          ('2013-08-31', 0.87, 2141),
                          ('2013-09-28', 0.14, 2142),
                          ('2013-10-25', 0.43, 2143),
                          ('2013-11-21', 0.73, 2144),
                          ('2013-12-19', 0.05, 2145),
                          ('2014-01-15', 0.38, 2146),
                          ('2014-02-11', 0.73, 2147)])
def test_carrington_rotation_number(date, day_fraction, rotation_number):
    assert_allclose(sun.carrington_rotation_number(Time(date, scale='tt') + day_fraction*u.day),
                    rotation_number, rtol=0, atol=2e-4)


@pytest.mark.parametrize("crot, julian_days",
                         [(1, 2398167.4),
                          (2, 2398194.6756),
                          (1860, 2448872.027509),
                          ])
def test_carrington_rotation_starttime(crot, julian_days):
    # Stated precision in the docstring is 0.11 seconds
    with warnings.catch_warnings():
        # Filter warnings caused by very old dates
        warnings.filterwarnings("ignore", category=ErfaWarning)
        assert_quantity_allclose(sun.carrington_rotation_time(crot).tt.jd * u.day,
                                 julian_days * u.day, atol=0.11*u.s)


@pytest.mark.parametrize("crot, longitude, crot_fractional",
                         [(2000, 360, 2000.0),
                          (2000.0, 270, 2000.25)
                          ])
def test_carrington_rotation_time_longitude(crot, longitude, crot_fractional):
    assert sun.carrington_rotation_time(crot*u.one, longitude*u.deg) == \
        sun.carrington_rotation_time(crot_fractional*u.one)


@pytest.mark.parametrize("crot, longitude, crot_fractional",
                         [
                             (np.array([2000, 2000]), np.array(
                                 [180, 90]), np.array([2000.5, 2000.75])),
                             (2000, np.array([180, 90]), np.array([2000.5, 2000.75])),
                             (np.array([2000, 2000]), 180, np.array([2000.5, 2000.5]))
                         ])
def test_carrington_rotation_time_longitude_numpy(crot, longitude, crot_fractional):
    assert all(sun.carrington_rotation_time(crot*u.one, longitude*u.deg) ==
               sun.carrington_rotation_time(crot_fractional*u.one))


@pytest.mark.parametrize("crot, longitude",
                         [
                             (2000, 0),
                             (2000, -10),
                             (2000, 400),
                             (2000.5, 180),
                         ])
def test_carrington_rotation_time_longitude_err(crot, longitude):
    with pytest.raises(ValueError):
        sun.carrington_rotation_time(crot*u.one, longitude*u.deg)


def test_carrington_rotation_roundtrip():
    t = Time('2010-1-1')
    crot = sun.carrington_rotation_number(t)
    t_roundtrip = sun.carrington_rotation_time(crot)
    dt = t - t_roundtrip
    # Stated precision in the docstring is 0.11 seconds
    assert_quantity_allclose(dt.to(u.s), 0*u.s, atol=0.11*u.s)


def test_carrington_rotation_str():
    # Check that by default a human parseable string is returned
    t = sun.carrington_rotation_time(2210)
    assert str(t) == '2018-10-26 20:48:16.137'



from sunpy.net import attr
from sunpy.net import attrs as a
from sunpy.time import parse_time


def test_time_subclass():
    class NewTime(a.Time):
        pass

    assert isinstance(NewTime("2020/01/01", "2020/01/02") &
                      a.Time("2020/02/02", "2020/02/03"), attr.AttrAnd)


def test_attrs_time():
    times = a.Time("2020/10/01T00:00", "2020/10/01T00:00")
    times.start == parse_time("2020/10/01T00:00")
    times.end == parse_time("2020/10/01T00:00")
